"""Somesy log configuration."""
import logging
from enum import Enum, auto
from typing import Optional

from rich.logging import RichHandler

logger = logging.getLogger("somesy")

VERBOSE: int = 15
"""Custom logging level between INFO and DEBUG."""


class SomesyLogLevel(Enum):
    """Somesy-specific log levels."""

    SILENT = auto()
    INFO = auto()
    VERBOSE = auto()
    DEBUG = auto()

    @staticmethod
    def from_flags(
        *,
        info: Optional[bool] = None,
        verbose: Optional[bool] = None,
        debug: Optional[bool] = None
    ):
        """Convert CLI/config flags into a log level."""
        if debug:
            return SomesyLogLevel.DEBUG
        elif verbose:
            return SomesyLogLevel.VERBOSE
        elif info:
            return SomesyLogLevel.INFO
        return SomesyLogLevel.SILENT

    @staticmethod
    def to_logging(lv):
        """Convert a somesy log level into a logging log level."""
        if lv == SomesyLogLevel.SILENT:
            return logging.WARNING
        if lv == SomesyLogLevel.INFO:
            return logging.INFO
        if lv == SomesyLogLevel.VERBOSE:
            return VERBOSE
        if lv == SomesyLogLevel.DEBUG:
            return logging.DEBUG


_log_level: Optional[SomesyLogLevel] = None


def get_log_level() -> Optional[SomesyLogLevel]:
    """Return current user-defined log level."""
    return _log_level


def set_log_level(log_level: SomesyLogLevel) -> None:
    """Set the current log level."""
    global _log_level
    # update current somesy log level
    _log_level = log_level
    # (re-)init logging (rich formatter config depends on passed log level)
    init_log()
    # set the current logging log level
    logger.setLevel(SomesyLogLevel.to_logging(log_level))


def init_log():
    """Initialize logging (add VERBOSE log level and Rich formatter)."""
    _add_verbose_level()
    _init_rich_handler(get_log_level())


# ----


def _add_verbose_level():
    """Add a VERBOSE level to logging, if not already existing."""
    if isinstance(logging.getLevelName("VERBOSE"), int):
        return  # nothing to do

    # add the new level, if not defined yet
    logging.addLevelName(level=VERBOSE, levelName="VERBOSE")
    logger.propagate = False

    def verbose_print(self, message, *args, **kwargs):
        """Verbose logging level print function."""
        if self.isEnabledFor(VERBOSE):
            self._log(VERBOSE, message.format(args), (), **kwargs)

    setattr(logging.Logger, "verbose", verbose_print)  # noqa: B010
    logging.basicConfig(
        format="%(message)s",
        datefmt="",
    )


_rich_handler = None


def _init_rich_handler(log_level):
    """(Re-)initialize rich logging handler, based on current log level."""
    global _rich_handler
    debug = log_level == SomesyLogLevel.DEBUG

    if _rich_handler is not None:  # remove old handler
        logger.removeHandler(_rich_handler)

    # create and add new handler (based on log level)
    _rich_handler = RichHandler(
        show_time=False,
        rich_tracebacks=True,
        show_level=debug,
        show_path=debug,
        tracebacks_show_locals=debug,
        markup=True,
    )
    logger.addHandler(_rich_handler)



from pathlib import Path

import pytest

from somesy.package_json.writer import PackageJSON


def test_load():
    not_exist_path = Path("reject/package.json")
    with pytest.raises(FileNotFoundError):
        PackageJSON(not_exist_path)

    file_path = Path("tests/package_json/data/package.json")
    PackageJSON(file_path)
    assert file_path.is_file()

    reject_path = Path("tests/package_json/data/reject/package.json")
    with pytest.raises(ValueError):
        PackageJSON(reject_path)


def test_sync(somesy, package_json):
    metadata = somesy.project
    package_json.sync(metadata)
    assert package_json.name == "testproject"
    assert package_json.version == "1.0.0"
    assert (
        package_json.description == "This is a test project for demonstration purposes."
    )



"""Pyproject models."""
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Set, Union

from packaging.version import parse as parse_version
from pydantic import (
    BaseModel,
    EmailStr,
    Field,
    HttpUrl,
    ValidationError,
    root_validator,
    validator,
)
from typing_extensions import Annotated

from somesy.core.models import LicenseEnum


class PoetryConfig(BaseModel):
    """Poetry configuration model."""

    name: Annotated[
        str,
        Field(regex=r"^[A-Za-z0-9]+([_-][A-Za-z0-9]+)*$", description="Package name"),
    ]
    version: Annotated[
        str,
        Field(
            regex=r"^\d+(\.\d+)*((a|b|rc)\d+)?(post\d+)?(dev\d+)?$",
            description="Package version",
        ),
    ]
    description: Annotated[str, Field(description="Package description")]
    license: Annotated[
        Optional[Union[LicenseEnum, List[LicenseEnum]]],
        Field(description="An SPDX license identifier."),
    ]
    authors: Annotated[Set[str], Field(description="Package authors")]
    maintainers: Annotated[Optional[Set[str]], Field(description="Package maintainers")]
    readme: Annotated[
        Optional[Union[Path, List[Path]]], Field(description="Package readme file(s)")
    ]
    homepage: Annotated[Optional[HttpUrl], Field(description="Package homepage")]
    repository: Annotated[Optional[HttpUrl], Field(description="Package repository")]
    documentation: Annotated[
        Optional[HttpUrl], Field(description="Package documentation page")
    ]
    keywords: Annotated[
        Optional[Set[str]], Field(description="Keywords that describe the package")
    ]
    classifiers: Annotated[Optional[List[str]], Field(description="pypi classifiers")]
    urls: Annotated[Optional[Dict[str, HttpUrl]], Field(description="Package URLs")]

    @validator("version")
    def validate_version(cls, v):
        """Validate version using PEP 440."""
        try:
            _ = parse_version(v)
        except ValueError:
            raise ValidationError("Invalid version")
        return v

    @validator("authors", "maintainers")
    def validate_email_format(cls, v):
        """Validate email format."""
        for author in v:
            if (
                not isinstance(author, str)
                or " " not in author
                or not EmailStr.validate(author.split(" ")[-1][1:-1])
            ):
                raise ValidationError("Invalid email format")
        return v

    @validator("readme")
    def validate_readme(cls, v):
        """Validate readme file(s) by checking whether files exist."""
        if type(v) is list:
            if any(not e.is_file() for e in v):
                raise ValidationError("Some file(s) do not exist")
        else:
            if not v.is_file():
                raise ValidationError("File does not exist")

    class Config:
        """Pydantic configuration."""

        use_enum_values = True


class ContentTypeEnum(Enum):
    """Content type enum for setuptools field file."""

    plain = "text/plain"
    rst = "text/x-rst"
    markdown = "text/markdown"


class File(BaseModel):
    """File model for setuptools."""

    file: Path
    content_type: Optional[ContentTypeEnum] = Field(alias="content-type")


class License(BaseModel):
    """License model for setuptools."""

    file: Optional[Path]
    text: Optional[LicenseEnum]

    class Config:
        """Pydantic configuration."""

        validate_assignment = True

    @root_validator(pre=True)
    def validate_xor(cls, values):
        """Validate that only one of file or text is set."""
        if sum([bool(v) for v in values.values()]) != 1:
            raise ValueError("Either file or text must be set.")
        return values


class STPerson(BaseModel):
    """Person model for setuptools."""

    name: Annotated[str, Field(min_length=1)]
    email: Annotated[str, Field(min_length=1)]


class URLs(BaseModel):
    """URLs model for setuptools."""

    homepage: Optional[HttpUrl] = None
    repository: Optional[HttpUrl] = None
    documentation: Optional[HttpUrl] = None
    changelog: Optional[HttpUrl] = None


class SetuptoolsConfig(BaseModel):
    """Setuptools input model. Required fields are name, version, description, and requires_python."""

    name: Annotated[str, Field(regex=r"^[A-Za-z0-9]+([_-][A-Za-z0-9]+)*$")]
    version: Annotated[
        str, Field(regex=r"^\d+(\.\d+)*((a|b|rc)\d+)?(post\d+)?(dev\d+)?$")
    ]
    description: str
    readme: Optional[Union[Path, List[Path], File]] = None
    license: Optional[Union[LicenseEnum, List[LicenseEnum]]] = Field(
        None, description="An SPDX license identifier."
    )
    authors: Optional[List[STPerson]]
    maintainers: Optional[List[STPerson]]
    keywords: Optional[Set[str]]
    classifiers: Optional[List[str]]
    urls: Optional[URLs]

    @validator("version")
    def validate_version(cls, v):
        """Validate version using PEP 440."""
        try:
            _ = parse_version(v)
        except ValueError:
            raise ValidationError("Invalid version")
        return v

    @validator("readme")
    def validate_readme(cls, v):
        """Validate readme file(s) by checking whether files exist."""
        if type(v) is list:
            if any(not e.is_file() for e in v):
                raise ValidationError("Some file(s) do not exist")
        elif type(v) is File:
            if not Path(v.file).is_file():
                raise ValidationError("File does not exist")
        else:
            if not v.is_file():
                raise ValidationError("File does not exist")

    @validator("authors", "maintainers")
    def validate_email_format(cls, v):
        """Validate email format."""
        for person in v:
            if person.email:
                if not EmailStr.validate(person.email):
                    raise ValidationError("Invalid email format")
        return v

    class Config:
        """Pydantic configuration."""

        use_enum_values = True



from pathlib import Path

import pytest

from somesy.core.core import discover_input
from somesy.core.models import ProjectMetadata, SomesyConfig, SomesyInput


@pytest.fixture
def somesy_metadata_only():
    return Path("tests/core/data/.somesy.toml")


@pytest.fixture
def somesy_with_config():
    return Path("tests/core/data/.somesy.with_config.toml")


def test_discover_input(tmp_path, monkeypatch: pytest.MonkeyPatch):
    # Test 1: input is is given and exists
    input_file = Path("tests/core/data/.somesy.toml")
    result = discover_input(input_file)
    assert result == input_file

    # Test 2: input is is given but does not exist
    input_file = Path("tests/core/data/.somesy2.toml")
    result = discover_input(input_file)
    assert result == Path(".somesy.toml")

    monkeypatch.chdir(tmp_path)
    input_file = Path("tests/core/data/.somesy2.toml")
    with pytest.raises(FileNotFoundError):
        discover_input(input_file)


def test_with_project_metadata(somesy_metadata_only):
    # valid somesy file
    metadata = SomesyInput.from_input_file(somesy_metadata_only).project
    assert isinstance(metadata, ProjectMetadata)
    assert metadata.name == "somesy"
    assert metadata.version == "0.1.0"


def test_with_extract_cli_config(somesy_with_config, somesy_metadata_only):
    # test with config exists
    config = SomesyInput.from_input_file(somesy_with_config).config
    assert isinstance(config, SomesyConfig)
    assert config.debug == True

    # test with config does not exist
    config = SomesyInput.from_input_file(somesy_metadata_only).config
    assert config is None


def test_somesy_input(somesy_metadata_only):
    metadata = SomesyInput.from_input_file(somesy_metadata_only).project
    assert isinstance(metadata, ProjectMetadata)
    assert metadata.name == "somesy"
    assert metadata.version == "0.1.0"


def test_somesy_input2(somesy_with_config, somesy_metadata_only):
    # test with config exists
    config = SomesyInput.from_input_file(somesy_with_config).config
    assert isinstance(config, SomesyConfig)
    assert config.debug == True

    # test with config does not exist
    config = SomesyInput.from_input_file(somesy_metadata_only).config
    assert config is None



"""Core somesy functions."""
import json
import logging
from pathlib import Path
from typing import Any, Dict, Optional

import tomlkit

logger = logging.getLogger("somesy")

INPUT_FILES_ORDERED = [".somesy.toml", "somesy.toml", "pyproject.toml", "package.json"]
"""Input files ordered by priority for discovery."""


def discover_input(input_file: Optional[Path] = None) -> Path:
    """Check given input file path. If not given, find somesy configuration file path from default list.

    Args:
        input_file: somesy configuration file path. Defaults to None.

    Raises:
        FileNotFoundError: Raised if no somesy input file found from cli input or the defaults.

    Returns:
        somesy configuration file path.
    """
    if input_file:
        if input_file.is_file():
            logger.info(f"Using provided file '{input_file}' as somesy input file.")
            return input_file
        else:
            msg = f"Passed file '{input_file}' does not exist. Searching for usable somesy input file..."
            logger.verbose(msg)

    for filename in INPUT_FILES_ORDERED:
        input_file = Path(filename)
        if input_file.is_file():
            try:
                get_input_content(input_file)
            except RuntimeError:
                continue

            msg = f"Using '{input_file}' as somesy input file."
            logger.verbose(msg)
            return input_file

    raise FileNotFoundError("No somesy input file found.")


def get_input_content(path: Path, *, no_unwrap: bool = False) -> Dict[str, Any]:
    """Read contents of a supported somesy input file.

    Given a path to a TOML file, this function reads the file and returns its content as a TOMLDocument object.
    The function checks if the file is a valid somesy input file by checking its name and content.

    Args:
        path (Path): path to the input file

    Returns:
        the content of the input file as a TOMLDocument object

    Raises:
        ValueError: if the input file is not a valid somesy input file or if the file is not a TOML file.
        RuntimeError: if the input file does not contain a somesy input section at expected key
    """
    logger.debug(f"Path {path}")
    # somesy.toml / .somesy.toml
    if path.suffix == ".toml" and "somesy" in path.name:
        with open(path, "r") as f:
            ret = tomlkit.load(f)
            return ret if no_unwrap else ret.unwrap()

    # pyproject.toml
    if path.suffix == ".toml" and "pyproject" in path.name:
        with open(path, "r") as f:
            input_content = tomlkit.load(f)
            if "tool" in input_content and "somesy" in input_content["tool"]:
                return input_content["tool"]["somesy"].unwrap()
            else:
                raise RuntimeError(
                    "No tool.somesy section found in pyproject.toml file!"
                )

    if path.suffix == ".json" and "package" in path.name:
        with open(path, "r") as f:
            input_content = json.load(f)
            if "somesy" in input_content:
                return input_content["somesy"]
            else:
                raise RuntimeError("No somesy section found in package.json file!")

    # no match:
    raise ValueError("Unsupported input file.")



import json
import sys

import pytest
import rdflib

from somesy.cff.writer import CFF
from somesy.codemeta import update_codemeta
from somesy.codemeta.utils import (
    _codemeta_context,
    _graph_from_cm_dict,
    _graph_from_cm_file,
    _localize_codemetapy_context,
)
from somesy.core.models import SomesyConfig

cm_dict = {
    "@context": list(_codemeta_context),
    "@type": "SoftwareSourceCode",
    "applicationCategory": "Software Development > Libraries > Application Frameworks",
    "audience": [
        {"@type": "Audience", "audienceType": "Developers"},
    ],
    "author": [
        {
            "@id": "https://orcid.org/0123-4567-8910-1112",
            "@type": "Person",
            "familyName": "Doe",
            "givenName": "Jane",
        },
    ],
    "codeRepository": "https://example.com/my-software",
    "description": "Amazing software",
    "identifier": "awesome-tool",
    "keywords": ["amazing", "software"],
    "license": "http://spdx.org/licenses/MIT",
    "name": "awesome-tool",
    "operatingSystem": "POSIX > Linux",
    "runtimePlatform": "Python 3",
    "url": "https://example.com/my-software",
    "version": "0.1.0",
}
num_triples = 20  # empirically observed from successful rdflib parse


def test_localized_graph():
    # no context -> do nothing
    assert _localize_codemetapy_context({}) == {}

    # unexpected context -> exception
    with pytest.raises(RuntimeError):
        _localize_codemetapy_context({"@context": "invalid"})

    # correct context -> replace
    ret = _localize_codemetapy_context({"@context": _codemeta_context})
    assert ret["@context"] != _codemeta_context


def rdflib_audit_hook(name: str, args) -> None:
    """An audit hook that blocks access when an attempt is made to open an URL.

    See https://rdflib.readthedocs.io/en/stable/_modules/examples/secure_with_audit.html#audit_hook
    """
    if name == "urllib.Request":
        raise PermissionError("Permission denied for URL")
    return None


def test_graph_from_dict():
    sys.addaudithook(rdflib_audit_hook)

    # test without localized graph
    # -> rdflib will attempt network access
    g = rdflib.Graph()
    with pytest.raises(PermissionError):
        g.parse(data=json.dumps(cm_dict), format="json-ld")

    # now use the localized graph with local context
    # -> no network access, but still able to load triples
    ret = _graph_from_cm_dict(cm_dict)
    assert len(ret) == num_triples


def test_graph_from_file(tmp_path):
    filepath = tmp_path / "my_codemeta.json"

    # no file -> None
    assert _graph_from_cm_file(filepath) is None
    # create file
    with open(filepath, "w") as f:
        json.dump(cm_dict, f)
    # file exists -> returns correct graph
    ret = _graph_from_cm_file(filepath)
    assert len(ret) == num_triples


def test_update_codemeta(tmp_path):
    file_path = tmp_path / "my_codemeta.json"

    cff_path = tmp_path / "CITATION.cff"
    cff = CFF(cff_path)
    cff.save()
    assert cff_path.is_file()

    # first time, no file -> codemeta.json is created
    conf = SomesyConfig(codemeta_file=file_path, cff_file=cff_path)
    assert update_codemeta(conf)

    # second time, no changes -> codemeta.json is not overwritten
    assert not update_codemeta(conf)

    cff.description = "Changed description"
    cff.save()
    # second time, changes -> codemeta.json is overwritten
    assert update_codemeta(conf)



from pathlib import Path

import pytest

from somesy.core.log import SomesyLogLevel, set_log_level
from somesy.core.models import SomesyInput
from somesy.package_json.writer import PackageJSON


@pytest.fixture(scope="session", autouse=True)
def init_somesy_logger():
    set_log_level(SomesyLogLevel.DEBUG)


@pytest.fixture
def create_poetry_file():
    def _create_poetry_file(pyproject_file: Path):
        # create pyproject file beforehand
        with open("tests/pyproject/data/pyproject.full.toml", "r") as f:
            pyproject_content = f.read()
            with open(pyproject_file, "w+") as f2:
                f2.write(pyproject_content)

    yield _create_poetry_file


@pytest.fixture
def create_setuptools_file():
    def _create_setuptools_file(pyproject_file: Path):
        # create pyproject file beforehand
        with open("tests/pyproject/data/pyproject.setuptools.toml", "r") as f:
            pyproject_content = f.read()
            with open(pyproject_file, "w+") as f2:
                f2.write(pyproject_content)

    yield _create_setuptools_file


@pytest.fixture
def create_somesy_metadata():
    def _create_somesy_metadata(somesy_file: Path):
        # create somesy file beforehand
        with open("tests/core/data/.somesy.toml", "r") as f:
            content = f.read()
            with open(somesy_file, "w+") as f2:
                f2.write(content)

    yield _create_somesy_metadata


@pytest.fixture
def create_somesy():
    def _create_somesy(somesy_file: Path):
        # create somesy file beforehand
        with open("tests/data/somesy.toml", "r") as f:
            content = f.read()
            with open(somesy_file, "w+") as f2:
                f2.write(content)

    yield _create_somesy


@pytest.fixture
def create_package_json():
    def _create_package_json(package_json_file: Path):
        # create somesy file beforehand
        with open("tests/data/package.json", "r") as f:
            content = f.read()
            with open(package_json_file, "w+") as f2:
                f2.write(content)

    yield _create_package_json


@pytest.fixture
def create_cff_file():
    def _create_cff_file(cff_file: Path):
        # create somesy file beforehand
        with open("tests/cff/data/CITATION.cff", "r") as f:
            content = f.read()
            with open(cff_file, "w+") as f2:
                f2.write(content)

    yield _create_cff_file


@pytest.fixture
def somesy() -> dict:
    return SomesyInput.from_input_file(Path("tests/data/somesy.toml"))


@pytest.fixture
def package_json() -> PackageJSON:
    return PackageJSON(Path("tests/data/package.json"))



import pytest

from somesy.commands.init_config import init_config


@pytest.fixture
def cli_options() -> dict:
    return {
        "no_sync_cff": False,
        "cff_file": "CITATION.cff",
        "no_sync_pyproject": False,
        "pyproject_file": "pyproject.toml",
        "sync_package_json": True,
        "package_json_file": "package.json",
        "show_info": False,
        "verbose": False,
        "debug": False,
    }


def test_init_config(
    tmp_path,
    create_somesy_metadata,
    create_somesy,
    create_poetry_file,
    cli_options,
):
    # load somesy file
    somesy_file = tmp_path / ".somesy.toml"
    create_somesy_metadata(somesy_file)
    options = dict(cli_options)
    options["debug"] = True
    init_config(somesy_file, options)
    assert "debug = true" in somesy_file.read_text()

    # load somesy file with config
    somesy_file = tmp_path / ".somesy.with_config.toml"
    create_somesy(somesy_file)
    options = dict(cli_options)
    options["show_info"] = True
    init_config(somesy_file, options)
    assert "show_info = true" in somesy_file.read_text()

    # load pyproject file
    pyproject_file = tmp_path / "pyproject.toml"
    create_poetry_file(pyproject_file)
    options = dict(cli_options)
    options["debug"] = True
    init_config(pyproject_file, options)
    assert "debug = true" in pyproject_file.read_text()



"""package.json validation models."""
import re
from typing import List, Optional, Union

from pydantic import AnyUrl, BaseModel, EmailStr, Field, ValidationError, validator
from typing_extensions import Annotated


class PackageAuthor(BaseModel):
    """Package author model."""

    name: Annotated[Optional[str], Field(description="Author name")]
    email: Annotated[Optional[EmailStr], Field(description="Author email")]
    url: Annotated[Optional[AnyUrl], Field(description="Author website or orcid page")]


class PackageRepository(BaseModel):
    """Package repository model."""

    type: Annotated[Optional[str], Field(description="Repository type")]
    url: Annotated[str, Field(description="Repository url")]


class PackageLicense(BaseModel):
    """Package license model."""

    type: Annotated[Optional[str], Field(description="License type")]
    url: Annotated[str, Field(description="License url")]


class PackageJsonConfig(BaseModel):
    """Package.json config model."""

    name: Annotated[str, Field(description="Package name")]
    version: Annotated[str, Field(description="Package version")]
    description: Annotated[Optional[str], Field(description="Package description")]
    author: Annotated[
        Optional[Union[str, PackageAuthor]], Field(description="Package author")
    ]
    maintainers: Annotated[
        Optional[List[Union[str, PackageAuthor]]],
        Field(description="Package maintainers"),
    ]
    contributors: Annotated[
        Optional[List[Union[str, PackageAuthor]]],
        Field(description="Package contributors"),
    ]
    license: Annotated[
        Optional[Union[str, PackageLicense]], Field(description="Package license")
    ]
    repository: Annotated[
        Optional[Union[PackageRepository, str]], Field(description="Package repository")
    ]
    homepage: Annotated[Optional[AnyUrl], Field(description="Package homepage")]
    keywords: Annotated[
        Optional[List[str]], Field(description="Keywords that describe the package")
    ]

    # convert package author to dict if it is a string
    @classmethod
    def convert_author(cls, author: str) -> PackageAuthor:
        """Convert author string to PackageAuthor model."""
        # parse author string to "name <email> (url)" format with regex
        author_regex = r"^(.*?)\s*(?:<([^>]+)>)?\s*(?:\(([^)]+)\))?$"
        author_match = re.match(author_regex, author)
        if not author_match:
            raise ValidationError(f"Invalid author format: {author}")
        author_name = author_match[1]
        author_email = author_match[2]
        author_url = author_match[3]

        return PackageAuthor(name=author_name, email=author_email, url=author_url)

    @validator("name")
    def validate_name(cls, v):
        """Validate package name."""
        pattern = r"^(@[a-z0-9-~][a-z0-9-._~]*\/)?[a-z0-9-~][a-z0-9-._~]*$"
        if re.match(pattern, v) is None:
            raise ValidationError("Invalid name")

        return v

    @validator("version")
    def validate_version(cls, v):
        """Validate package version."""
        # pattern for npm version
        pattern = r"^(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)\.(?:0|[1-9]\d*)(?:-(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*)?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"
        if re.match(pattern, v) is None:
            raise ValidationError("Invalid version")
        return v

    @validator("author")
    def validate_author(cls, v):
        """Validate package author."""
        return cls.convert_author(v) if isinstance(v, str) else v

    @validator("maintainers", "contributors")
    def validate_people(cls, v):
        """Validate package maintainers and contributors."""
        people = []
        for p in v:
            if isinstance(p, str):
                people.append(cls.convert_author(p))
            else:
                people.append(p)
        return people

    class Config:
        """Pydantic config."""

        allow_population_by_field_name = True



"""Set config files for somesy."""
import logging
from pathlib import Path

import typer

from somesy.commands import init_config
from somesy.core.core import discover_input
from somesy.core.log import SomesyLogLevel, set_log_level

from .util import wrap_exceptions

logger = logging.getLogger("somesy")
app = typer.Typer()


@app.command()
@wrap_exceptions
def config():
    """Set CLI configs for somesy."""
    # check if input file exists, if not, try to find it from default list
    input_file_default = discover_input()

    # prompt for inputs
    input_file = typer.prompt("Input file path", default=input_file_default)
    input_file = Path(input_file)
    options = {
        "input_file": input_file,
        "no_sync_cff": not typer.confirm(
            "Do you want to sync to a CFF file?", default=True
        ),
    }
    cff_file = typer.prompt("CFF file path", default="CITATION.cff")
    if cff_file is not None or cff_file != "":
        options["cff_file"] = cff_file

    options["no_sync_pyproject"] = not typer.confirm(
        "Do you want to sync to a pyproject.toml file?", default=True
    )

    pyproject_file = typer.prompt("pyproject.toml file path", default="pyproject.toml")
    if pyproject_file is not None or pyproject_file != "":
        options["pyproject_file"] = pyproject_file

    options["sync_package_json"] = typer.confirm(
        "Do you want to sync to a package.json file?", default=False
    )
    package_json_file = typer.prompt("package.json file path", default="package.json")
    if package_json_file is not None or package_json_file != "":
        options["package_json_file"] = package_json_file

    options["no_sync_codemeta"] = not typer.confirm(
        "Do you want to sync to a codemeta.json file?", default=True
    )
    codemeta_file = typer.prompt("codemeta.json file path", default="codemeta.json")
    if codemeta_file is not None or codemeta_file != "":
        options["codemeta_file"] = codemeta_file

    options["show_info"] = typer.confirm(
        "Do you want to show info about the sync process?"
    )
    options["verbose"] = typer.confirm("Do you want to show verbose logs?")
    options["debug"] = typer.confirm("Do you want to show debug logs?")

    set_log_level(
        SomesyLogLevel.from_flags(
            debug=options["debug"],
            verbose=options["verbose"],
            info=options["show_info"],
        )
    )

    logger.debug(f"CLI options entered: {options}")

    init_config(input_file, options)
    logger.info(
        f"[bold green]Input file is updated/created at {input_file}[/bold green]"
    )



"""package.json parser and saver."""
import json
import logging
from collections import OrderedDict
from pathlib import Path
from typing import List, Optional

from rich.pretty import pretty_repr

from somesy.core.models import Person, ProjectMetadata
from somesy.core.writer import ProjectMetadataWriter
from somesy.package_json.models import PackageJsonConfig

logger = logging.getLogger("somesy")


class PackageJSON(ProjectMetadataWriter):
    """package.json parser and saver."""

    def __init__(
        self,
        path: Path,
    ):
        """package.json parser.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        mappings = {
            "authors": ["author"],
        }
        super().__init__(path, create_if_not_exists=False, direct_mappings=mappings)

    @property
    def authors(self):
        """Return the only author of the package.json file as list."""
        return [self._get_property(self._get_key("authors"))]

    @authors.setter
    def authors(self, authors: List[Person]) -> None:
        """Set the authors of the project."""
        authors = self._from_person(authors[0])
        self._set_property(self._get_key("authors"), authors)

    @property
    def contributors(self):
        """Return the contributors of the package.json file."""
        return self._get_property(self._get_key("contributors"))

    @contributors.setter
    def contributors(self, contributors: List[Person]) -> None:
        """Set the contributors of the project."""
        contributors = [self._from_person(c) for c in contributors]
        self._set_property(self._get_key("contributors"), contributors)

    def _load(self) -> None:
        """Load package.json file."""
        with self.path.open() as f:
            self._data = json.load(f, object_pairs_hook=OrderedDict)

    def _validate(self) -> None:
        """Validate package.json content using pydantic class."""
        config = dict(self._get_property([]))
        logger.debug(
            f"Validating config using {PackageJsonConfig.__name__}: {pretty_repr(config)}"
        )
        PackageJsonConfig(**config)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the package.json file."""
        path = path or self.path
        logger.debug(f"Saving package.json to {path}")

        with path.open("w") as f:
            # package.json indentation is 2 spaces
            json.dump(self._data, f, indent=2)

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to package.json dict for person format."""
        person_dict = {"name": person.full_name}
        if person.email:
            person_dict["email"] = person.email
        if person.orcid:
            person_dict["url"] = person.orcid
        return person_dict

    @staticmethod
    def _to_person(person) -> Person:
        """Convert package.json dict or str for person format to project metadata person object."""
        if isinstance(person, str):
            # parse from package.json format
            person = PackageJsonConfig.convert_author(person).dict(exclude_none=True)

        names = list(map(lambda s: s.strip(), person["name"].split()))
        person_obj = {
            "given-names": " ".join(names[:-1]),
            "family-names": names[-1],
        }
        if "email" in person:
            person_obj["email"] = person["email"].strip()
        if "url" in person:
            person_obj["orcid"] = person["url"].strip()
        return Person(**person_obj)

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync package.json with project metadata.

        Use existing sync function from ProjectMetadataWriter but update repository and contributors.
        """
        super().sync(metadata)
        self.contributors = self._sync_person_list(self.contributors, metadata.people)
        if metadata.repository:
            self.repository = {"type": "git", "url": metadata.repository}



"""Sync selected metadata files with given input file."""
import logging
from pathlib import Path

from rich.pretty import pretty_repr

from somesy.cff.writer import CFF
from somesy.codemeta import update_codemeta
from somesy.core.models import ProjectMetadata, SomesyConfig, SomesyInput
from somesy.package_json.writer import PackageJSON
from somesy.pyproject.writer import Pyproject

logger = logging.getLogger("somesy")


def sync(somesy_input: SomesyInput):
    """Sync selected metadata files with given input file."""
    conf, metadata = somesy_input.config, somesy_input.project

    logger.debug(
        f"Project metadata: {pretty_repr(metadata.dict(exclude_defaults=True))}"
    )

    # update these only if they exist:

    if conf.pyproject_file.is_file() and not conf.no_sync_pyproject:
        _sync_python(metadata, conf.pyproject_file)

    if conf.package_json_file.is_file() and not conf.no_sync_package_json:
        _sync_package_json(metadata, conf.package_json_file)

    # create these by default if they are missing:

    if not conf.no_sync_cff:
        _sync_cff(metadata, conf.cff_file)

    # NOTE: codemeta should always be last, it uses (some of) the other targets
    if not conf.no_sync_codemeta:
        _sync_codemeta(conf)


def _sync_python(
    metadata: ProjectMetadata,
    pyproject_file: Path,
):
    """Sync pyproject.toml file using project metadata.

    Args:
        metadata (ProjectMetadata): project metadata to sync pyproject.toml file.
        pyproject_file (Path, optional): pyproject file to read project metadata from.
    """
    logger.verbose("Loading pyproject.toml file.")
    pyproject = Pyproject(pyproject_file)
    logger.verbose("Syncing pyproject.toml file.")
    pyproject.sync(metadata)
    pyproject.save()
    logger.verbose("Saved synced pyproject.toml file.\n")


def _sync_cff(
    metadata: ProjectMetadata,
    cff_file: Path,
):
    """Sync CITATION.cff file using project metadata.

    Args:
        metadata (ProjectMetadata): project metadata to sync pyproject.toml file.
        cff_file (Path, optional): CFF file path if wanted to be synced. Defaults to None.
    """
    logger.verbose("Loading CITATION.cff file.")
    cff = CFF(cff_file)
    logger.verbose("Syncing CITATION.cff file.")
    cff.sync(metadata)
    cff.save()
    logger.verbose("Saved synced CITATION.cff file.\n")


def _sync_package_json(
    metadata: ProjectMetadata,
    package_json_file: Path,
):
    """Sync package.json file using project metadata.

    Args:
        metadata (ProjectMetadata): project metadata to sync pyproject.toml file.
        package_json_file (Path, optional): package.json file path if wanted to be synced. Defaults to None.
    """
    logger.verbose("Loading package.json file.")
    package_json = PackageJSON(package_json_file)
    logger.verbose("Syncing package.json file.")
    package_json.sync(metadata)
    package_json.save()
    logger.verbose("Saved synced package.json file.\n")


def _sync_codemeta(conf: SomesyConfig):
    logger.verbose("Updating codemeta.json file.")
    if update_codemeta(conf):
        logger.verbose(f"New codemeta graph written to {conf.codemeta_file}.")
    else:
        logger.verbose(f"Codemeta graph unchanged, keeping old {conf.codemeta_file}.")



from typing import OrderedDict

from somesy.cff.writer import CFF
from somesy.commands.sync import sync
from somesy.core.models import Person, ProjectMetadata, SomesyConfig, SomesyInput
from somesy.package_json.writer import PackageJSON
from somesy.pyproject.writer import Pyproject


def test_sync(tmp_path, create_poetry_file, create_package_json, create_cff_file):
    # Create a temporary pyproject.toml file
    pyproject_file = tmp_path / "pyproject.toml"
    create_poetry_file(pyproject_file)
    pyproject = Pyproject(pyproject_file)

    # Create a temporary CITATION.cff file
    cff_file = tmp_path / "CITATION.cff"
    create_cff_file(cff_file)
    cff = CFF(cff_file)

    # Create a temporary package.json file
    package_json_file = tmp_path / "package.json"
    create_package_json(package_json_file)
    package_json = PackageJSON(package_json_file)

    # Create a SomesyInput object with some metadata
    metadata = ProjectMetadata(
        name="test-project",
        version="0.1.0",
        description="A test project",
        people=[
            Person(
                given_names="Alice", family_names="Joe", email="a@a.aa", author=True
            ),
            Person(given_names="Bob", family_names="Joe", email="b@b.bb"),
        ],
        license="MIT",
    )
    conf = SomesyConfig(
        pyproject_file=pyproject_file,
        cff_file=cff_file,
        no_sync_package_json=False,
        package_json_file=package_json_file,
        no_sync_codemeta=True,
    )
    somesy_input = SomesyInput(config=conf, project=metadata)

    # Call the sync function
    sync(
        somesy_input,
    )

    # Check that the pyproject.toml file was synced
    pyproject = Pyproject(pyproject_file)
    assert pyproject.name == "test-project"
    assert pyproject.version == "0.1.0"
    assert pyproject.description == "A test project"
    assert pyproject.authors == ["Alice Joe <a@a.aa>"]
    assert pyproject.license == "MIT"

    # Check that the CITATION.cff file was synced
    cff = CFF(cff_file)
    assert cff.name == "test-project"
    assert cff.version == "0.1.0"
    assert cff.description == "A test project"
    assert cff.authors == [
        {"given-names": "Alice", "email": "a@a.aa", "family-names": "Joe"}
    ]
    assert cff.license == "MIT"

    # Check that the package.json file was synced
    package_json = PackageJSON(package_json_file)
    assert package_json.name == "test-project"
    assert package_json.version == "0.1.0"
    assert package_json.description == "A test project"
    assert package_json.authors == [
        OrderedDict([("name", "Alice Joe"), ("email", "a@a.aa")])
    ]
    assert package_json.license == "MIT"



from pathlib import Path

from typer.testing import CliRunner

from somesy.core import core
from somesy.main import app

runner = CliRunner()


def test_app_sync(tmp_path, create_poetry_file, mocker):
    input_file = Path("tests/core/data/.somesy.with_config.toml")
    cff_file = tmp_path / "CITATION.cff"
    pyproject_file = tmp_path / "pyproject.toml"
    codemeta_file = tmp_path / "codemeta.json"

    # test sync without output files
    result = runner.invoke(
        app,
        [
            "sync",
            "-i",
            str(input_file),
            "--no-sync-pyproject",
            "--no-sync-cff",
            "--no-sync-codemeta",
        ],
    )
    assert result.exit_code == 1
    assert "nothing to do" in result.stdout

    # create pyproject file beforehand
    create_poetry_file(pyproject_file)

    # test sync with output files
    result = runner.invoke(
        app,
        [
            "-vvv",
            "sync",
            "-i",
            str(input_file),
            "-c",
            cff_file,
            "-p",
            pyproject_file,
            "-m",
            codemeta_file,
        ],
    )
    print(result.output)
    assert result.exit_code == 0
    assert "Metadata synchronization completed." in result.stdout
    assert not cff_file.is_file()  # disabled in input_file!
    assert pyproject_file.is_file()
    assert codemeta_file.is_file()

    # create an error in the input file
    mocker.patch.object(core, "INPUT_FILES_ORDERED", [])
    input_file_reject = Path("tests/core/data/.somesy2.toml")
    result = runner.invoke(
        app,
        [
            "-vvv",
            "sync",
            "-i",
            str(input_file_reject),
        ],
    )
    assert result.exit_code == 1
    assert "No somesy input file found." in result.stdout



"""Pyproject writers for setuptools and poetry."""
import logging
import re
from pathlib import Path
from typing import Any, List, Optional, Union

import tomlkit
import wrapt
from rich.pretty import pretty_repr
from tomlkit import load

from somesy.core.models import Person
from somesy.core.writer import ProjectMetadataWriter

from .models import PoetryConfig, SetuptoolsConfig

logger = logging.getLogger("somesy")


class PyprojectCommon(ProjectMetadataWriter):
    """Poetry config file handler parsed from pyproject.toml."""

    def __init__(
        self, path: Path, *, section: List[str], model_cls, direct_mappings=None
    ):
        """Poetry config file handler parsed from pyproject.toml.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        self._model_cls = model_cls
        self._section = section
        super().__init__(
            path, create_if_not_exists=False, direct_mappings=direct_mappings or {}
        )

    def _load(self) -> None:
        """Load pyproject.toml file."""
        with open(self.path) as f:
            self._data = tomlkit.load(f)

    def _validate(self) -> None:
        """Validate poetry config using pydantic class.

        In order to preserve toml comments and structure, tomlkit library is used.
        Pydantic class only used for validation.
        """
        config = dict(self._get_property([]))
        logger.debug(
            f"Validating config using {self._model_cls.__name__}: {pretty_repr(config)}"
        )
        self._model_cls(**config)

    def save(self, path: Optional[Path] = None) -> None:
        """Save the pyproject file."""
        path = path or self.path
        with open(path, "w") as f:
            tomlkit.dump(self._data, f)

    def _get_property(self, key: Union[str, List[str]]) -> Optional[Any]:
        """Get a property from the pyproject.toml file."""
        key_path = [key] if isinstance(key, str) else key
        full_path = self._section + key_path
        return super()._get_property(full_path)

    def _set_property(self, key: Union[str, List[str]], value: Any) -> None:
        """Set a property in the pyproject.toml file."""
        key_path = [key] if isinstance(key, str) else key
        # get the tomlkit object of the section
        dat = self._get_property([])
        # dig down, create missing nested objects on the fly
        curr = dat
        for key in key_path[:-1]:
            if key not in curr:
                curr.add(key, tomlkit.table())
            curr = curr[key]
        curr[key_path[-1]] = value


class Poetry(PyprojectCommon):
    """Poetry config file handler parsed from pyproject.toml."""

    def __init__(self, path: Path):
        """Poetry config file handler parsed from pyproject.toml.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        super().__init__(path, section=["tool", "poetry"], model_cls=PoetryConfig)

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to poetry string for person format "full name <email>."""
        return f"{person.full_name} <{person.email}>"

    @staticmethod
    def _to_person(person_obj: str) -> Person:
        """Parse poetry person string to a Person."""
        m = re.match(r"\s*([^<]+)<([^>]+)>", person_obj)
        names, mail = (
            list(map(lambda s: s.strip(), m.group(1).split())),
            m.group(2).strip(),
        )
        # NOTE: for our purposes, does not matter what are given or family names,
        # we only compare on full_name anyway.
        return Person(
            **{
                "given-names": " ".join(names[:-1]),
                "family-names": names[-1],
                "email": mail,
            }
        )


class SetupTools(PyprojectCommon):
    """Setuptools config file handler parsed from setup.cfg."""

    def __init__(self, path: Path):
        """Setuptools config file handler parsed from pyproject.toml.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        section = ["project"]
        mappings = {
            "homepage": ["urls", "homepage"],
            "repository": ["urls", "repository"],
        }
        super().__init__(
            path, section=section, direct_mappings=mappings, model_cls=SetuptoolsConfig
        )

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to setuptools dict for person format."""
        return {"name": person.full_name, "email": person.email}

    @staticmethod
    def _to_person(person_obj) -> Person:
        """Parse setuptools person string to a Person."""
        # NOTE: for our purposes, does not matter what are given or family names,
        # we only compare on full_name anyway.
        names = list(map(lambda s: s.strip(), person_obj["name"].split()))
        return Person(
            **{
                "given-names": " ".join(names[:-1]),
                "family-names": names[-1],
                "email": person_obj["email"].strip(),
            }
        )


# ----


class Pyproject(wrapt.ObjectProxy):
    """Class for syncing pyproject file with other metadata files."""

    __wrapped__: Union[SetupTools, Poetry]

    def __init__(self, path: Path):
        """Pyproject wrapper class. Wraps either setuptools or poetry.

        Args:
            path (Path): Path to pyproject.toml file.

        Raises:
            FileNotFoundError: Raised when pyproject.toml file is not found.
            ValueError: Neither project nor tool.poetry object is found in pyproject.toml file.
        """
        data = None
        if not path.is_file():
            raise FileNotFoundError(f"pyproject file {path} not found")

        with open(path, "r") as f:
            data = load(f)

        # inspect file to pick suitable project metadata writer
        if "project" in data:
            logger.verbose("Found setuptools-based metadata in pyproject.toml")
            self.__wrapped__ = SetupTools(path)
        elif "tool" in data and "poetry" in data["tool"]:
            logger.verbose("Found poetry-based metadata in pyproject.toml")
            self.__wrapped__ = Poetry(path)
        else:
            msg = "The pyproject.toml file is ambiguous, either add a [project] or [tool.poetry] section"
            raise ValueError(msg)

        super().__init__(self.__wrapped__)



"""Types and enums used in the somesy models."""
from enum import Enum


class LicenseEnum(Enum):
    """SPDX license identifiers."""

    field_0BSD = "0BSD"
    AAL = "AAL"
    Abstyles = "Abstyles"
    Adobe_2006 = "Adobe-2006"
    Adobe_Glyph = "Adobe-Glyph"
    ADSL = "ADSL"
    AFL_1_1 = "AFL-1.1"
    AFL_1_2 = "AFL-1.2"
    AFL_2_0 = "AFL-2.0"
    AFL_2_1 = "AFL-2.1"
    AFL_3_0 = "AFL-3.0"
    Afmparse = "Afmparse"
    AGPL_1_0 = "AGPL-1.0"
    AGPL_1_0_only = "AGPL-1.0-only"
    AGPL_1_0_or_later = "AGPL-1.0-or-later"
    AGPL_3_0 = "AGPL-3.0"
    AGPL_3_0_only = "AGPL-3.0-only"
    AGPL_3_0_or_later = "AGPL-3.0-or-later"
    Aladdin = "Aladdin"
    AMDPLPA = "AMDPLPA"
    AML = "AML"
    AMPAS = "AMPAS"
    ANTLR_PD = "ANTLR-PD"
    ANTLR_PD_fallback = "ANTLR-PD-fallback"
    Apache_1_0 = "Apache-1.0"
    Apache_1_1 = "Apache-1.1"
    Apache_2_0 = "Apache-2.0"
    APAFML = "APAFML"
    APL_1_0 = "APL-1.0"
    APSL_1_0 = "APSL-1.0"
    APSL_1_1 = "APSL-1.1"
    APSL_1_2 = "APSL-1.2"
    APSL_2_0 = "APSL-2.0"
    Artistic_1_0 = "Artistic-1.0"
    Artistic_1_0_cl8 = "Artistic-1.0-cl8"
    Artistic_1_0_Perl = "Artistic-1.0-Perl"
    Artistic_2_0 = "Artistic-2.0"
    Bahyph = "Bahyph"
    Barr = "Barr"
    Beerware = "Beerware"
    BitTorrent_1_0 = "BitTorrent-1.0"
    BitTorrent_1_1 = "BitTorrent-1.1"
    blessing = "blessing"
    BlueOak_1_0_0 = "BlueOak-1.0.0"
    Borceux = "Borceux"
    BSD_1_Clause = "BSD-1-Clause"
    BSD_2_Clause = "BSD-2-Clause"
    BSD_2_Clause_FreeBSD = "BSD-2-Clause-FreeBSD"
    BSD_2_Clause_NetBSD = "BSD-2-Clause-NetBSD"
    BSD_2_Clause_Patent = "BSD-2-Clause-Patent"
    BSD_2_Clause_Views = "BSD-2-Clause-Views"
    BSD_3_Clause = "BSD-3-Clause"
    BSD_3_Clause_Attribution = "BSD-3-Clause-Attribution"
    BSD_3_Clause_Clear = "BSD-3-Clause-Clear"
    BSD_3_Clause_LBNL = "BSD-3-Clause-LBNL"
    BSD_3_Clause_Modification = "BSD-3-Clause-Modification"
    BSD_3_Clause_No_Nuclear_License = "BSD-3-Clause-No-Nuclear-License"
    BSD_3_Clause_No_Nuclear_License_2014 = "BSD-3-Clause-No-Nuclear-License-2014"
    BSD_3_Clause_No_Nuclear_Warranty = "BSD-3-Clause-No-Nuclear-Warranty"
    BSD_3_Clause_Open_MPI = "BSD-3-Clause-Open-MPI"
    BSD_4_Clause = "BSD-4-Clause"
    BSD_4_Clause_Shortened = "BSD-4-Clause-Shortened"
    BSD_4_Clause_UC = "BSD-4-Clause-UC"
    BSD_Protection = "BSD-Protection"
    BSD_Source_Code = "BSD-Source-Code"
    BSL_1_0 = "BSL-1.0"
    BUSL_1_1 = "BUSL-1.1"
    bzip2_1_0_5 = "bzip2-1.0.5"
    bzip2_1_0_6 = "bzip2-1.0.6"
    C_UDA_1_0 = "C-UDA-1.0"
    CAL_1_0 = "CAL-1.0"
    CAL_1_0_Combined_Work_Exception = "CAL-1.0-Combined-Work-Exception"
    Caldera = "Caldera"
    CATOSL_1_1 = "CATOSL-1.1"
    CC_BY_1_0 = "CC-BY-1.0"
    CC_BY_2_0 = "CC-BY-2.0"
    CC_BY_2_5 = "CC-BY-2.5"
    CC_BY_3_0 = "CC-BY-3.0"
    CC_BY_3_0_AT = "CC-BY-3.0-AT"
    CC_BY_3_0_US = "CC-BY-3.0-US"
    CC_BY_4_0 = "CC-BY-4.0"
    CC_BY_NC_1_0 = "CC-BY-NC-1.0"
    CC_BY_NC_2_0 = "CC-BY-NC-2.0"
    CC_BY_NC_2_5 = "CC-BY-NC-2.5"
    CC_BY_NC_3_0 = "CC-BY-NC-3.0"
    CC_BY_NC_4_0 = "CC-BY-NC-4.0"
    CC_BY_NC_ND_1_0 = "CC-BY-NC-ND-1.0"
    CC_BY_NC_ND_2_0 = "CC-BY-NC-ND-2.0"
    CC_BY_NC_ND_2_5 = "CC-BY-NC-ND-2.5"
    CC_BY_NC_ND_3_0 = "CC-BY-NC-ND-3.0"
    CC_BY_NC_ND_3_0_IGO = "CC-BY-NC-ND-3.0-IGO"
    CC_BY_NC_ND_4_0 = "CC-BY-NC-ND-4.0"
    CC_BY_NC_SA_1_0 = "CC-BY-NC-SA-1.0"
    CC_BY_NC_SA_2_0 = "CC-BY-NC-SA-2.0"
    CC_BY_NC_SA_2_5 = "CC-BY-NC-SA-2.5"
    CC_BY_NC_SA_3_0 = "CC-BY-NC-SA-3.0"
    CC_BY_NC_SA_4_0 = "CC-BY-NC-SA-4.0"
    CC_BY_ND_1_0 = "CC-BY-ND-1.0"
    CC_BY_ND_2_0 = "CC-BY-ND-2.0"
    CC_BY_ND_2_5 = "CC-BY-ND-2.5"
    CC_BY_ND_3_0 = "CC-BY-ND-3.0"
    CC_BY_ND_4_0 = "CC-BY-ND-4.0"
    CC_BY_SA_1_0 = "CC-BY-SA-1.0"
    CC_BY_SA_2_0 = "CC-BY-SA-2.0"
    CC_BY_SA_2_0_UK = "CC-BY-SA-2.0-UK"
    CC_BY_SA_2_1_JP = "CC-BY-SA-2.1-JP"
    CC_BY_SA_2_5 = "CC-BY-SA-2.5"
    CC_BY_SA_3_0 = "CC-BY-SA-3.0"
    CC_BY_SA_3_0_AT = "CC-BY-SA-3.0-AT"
    CC_BY_SA_4_0 = "CC-BY-SA-4.0"
    CC_PDDC = "CC-PDDC"
    CC0_1_0 = "CC0-1.0"
    CDDL_1_0 = "CDDL-1.0"
    CDDL_1_1 = "CDDL-1.1"
    CDL_1_0 = "CDL-1.0"
    CDLA_Permissive_1_0 = "CDLA-Permissive-1.0"
    CDLA_Sharing_1_0 = "CDLA-Sharing-1.0"
    CECILL_1_0 = "CECILL-1.0"
    CECILL_1_1 = "CECILL-1.1"
    CECILL_2_0 = "CECILL-2.0"
    CECILL_2_1 = "CECILL-2.1"
    CECILL_B = "CECILL-B"
    CECILL_C = "CECILL-C"
    CERN_OHL_1_1 = "CERN-OHL-1.1"
    CERN_OHL_1_2 = "CERN-OHL-1.2"
    CERN_OHL_P_2_0 = "CERN-OHL-P-2.0"
    CERN_OHL_S_2_0 = "CERN-OHL-S-2.0"
    CERN_OHL_W_2_0 = "CERN-OHL-W-2.0"
    ClArtistic = "ClArtistic"
    CNRI_Jython = "CNRI-Jython"
    CNRI_Python = "CNRI-Python"
    CNRI_Python_GPL_Compatible = "CNRI-Python-GPL-Compatible"
    Condor_1_1 = "Condor-1.1"
    copyleft_next_0_3_0 = "copyleft-next-0.3.0"
    copyleft_next_0_3_1 = "copyleft-next-0.3.1"
    CPAL_1_0 = "CPAL-1.0"
    CPL_1_0 = "CPL-1.0"
    CPOL_1_02 = "CPOL-1.02"
    Crossword = "Crossword"
    CrystalStacker = "CrystalStacker"
    CUA_OPL_1_0 = "CUA-OPL-1.0"
    Cube = "Cube"
    curl = "curl"
    D_FSL_1_0 = "D-FSL-1.0"
    diffmark = "diffmark"
    DOC = "DOC"
    Dotseqn = "Dotseqn"
    DRL_1_0 = "DRL-1.0"
    DSDP = "DSDP"
    dvipdfm = "dvipdfm"
    ECL_1_0 = "ECL-1.0"
    ECL_2_0 = "ECL-2.0"
    eCos_2_0 = "eCos-2.0"
    EFL_1_0 = "EFL-1.0"
    EFL_2_0 = "EFL-2.0"
    eGenix = "eGenix"
    Entessa = "Entessa"
    EPICS = "EPICS"
    EPL_1_0 = "EPL-1.0"
    EPL_2_0 = "EPL-2.0"
    ErlPL_1_1 = "ErlPL-1.1"
    etalab_2_0 = "etalab-2.0"
    EUDatagrid = "EUDatagrid"
    EUPL_1_0 = "EUPL-1.0"
    EUPL_1_1 = "EUPL-1.1"
    EUPL_1_2 = "EUPL-1.2"
    Eurosym = "Eurosym"
    Fair = "Fair"
    Frameworx_1_0 = "Frameworx-1.0"
    FreeBSD_DOC = "FreeBSD-DOC"
    FreeImage = "FreeImage"
    FSFAP = "FSFAP"
    FSFUL = "FSFUL"
    FSFULLR = "FSFULLR"
    FTL = "FTL"
    GD = "GD"
    GFDL_1_1 = "GFDL-1.1"
    GFDL_1_1_invariants_only = "GFDL-1.1-invariants-only"
    GFDL_1_1_invariants_or_later = "GFDL-1.1-invariants-or-later"
    GFDL_1_1_no_invariants_only = "GFDL-1.1-no-invariants-only"
    GFDL_1_1_no_invariants_or_later = "GFDL-1.1-no-invariants-or-later"
    GFDL_1_1_only = "GFDL-1.1-only"
    GFDL_1_1_or_later = "GFDL-1.1-or-later"
    GFDL_1_2 = "GFDL-1.2"
    GFDL_1_2_invariants_only = "GFDL-1.2-invariants-only"
    GFDL_1_2_invariants_or_later = "GFDL-1.2-invariants-or-later"
    GFDL_1_2_no_invariants_only = "GFDL-1.2-no-invariants-only"
    GFDL_1_2_no_invariants_or_later = "GFDL-1.2-no-invariants-or-later"
    GFDL_1_2_only = "GFDL-1.2-only"
    GFDL_1_2_or_later = "GFDL-1.2-or-later"
    GFDL_1_3 = "GFDL-1.3"
    GFDL_1_3_invariants_only = "GFDL-1.3-invariants-only"
    GFDL_1_3_invariants_or_later = "GFDL-1.3-invariants-or-later"
    GFDL_1_3_no_invariants_only = "GFDL-1.3-no-invariants-only"
    GFDL_1_3_no_invariants_or_later = "GFDL-1.3-no-invariants-or-later"
    GFDL_1_3_only = "GFDL-1.3-only"
    GFDL_1_3_or_later = "GFDL-1.3-or-later"
    Giftware = "Giftware"
    GL2PS = "GL2PS"
    Glide = "Glide"
    Glulxe = "Glulxe"
    GLWTPL = "GLWTPL"
    gnuplot = "gnuplot"
    GPL_1_0 = "GPL-1.0"
    GPL_1_0_only = "GPL-1.0-only"
    GPL_1_0_or_later = "GPL-1.0-or-later"
    GPL_1_0_ = "GPL-1.0+"
    GPL_2_0 = "GPL-2.0"
    GPL_2_0_only = "GPL-2.0-only"
    GPL_2_0_or_later = "GPL-2.0-or-later"
    GPL_2_0_with_autoconf_exception = "GPL-2.0-with-autoconf-exception"
    GPL_2_0_with_bison_exception = "GPL-2.0-with-bison-exception"
    GPL_2_0_with_classpath_exception = "GPL-2.0-with-classpath-exception"
    GPL_2_0_with_font_exception = "GPL-2.0-with-font-exception"
    GPL_2_0_with_GCC_exception = "GPL-2.0-with-GCC-exception"
    GPL_2_0_ = "GPL-2.0+"
    GPL_3_0 = "GPL-3.0"
    GPL_3_0_only = "GPL-3.0-only"
    GPL_3_0_or_later = "GPL-3.0-or-later"
    GPL_3_0_with_autoconf_exception = "GPL-3.0-with-autoconf-exception"
    GPL_3_0_with_GCC_exception = "GPL-3.0-with-GCC-exception"
    GPL_3_0_ = "GPL-3.0+"
    gSOAP_1_3b = "gSOAP-1.3b"
    HaskellReport = "HaskellReport"
    Hippocratic_2_1 = "Hippocratic-2.1"
    HPND = "HPND"
    HPND_sell_variant = "HPND-sell-variant"
    HTMLTIDY = "HTMLTIDY"
    IBM_pibs = "IBM-pibs"
    ICU = "ICU"
    IJG = "IJG"
    ImageMagick = "ImageMagick"
    iMatix = "iMatix"
    Imlib2 = "Imlib2"
    Info_ZIP = "Info-ZIP"
    Intel = "Intel"
    Intel_ACPI = "Intel-ACPI"
    Interbase_1_0 = "Interbase-1.0"
    IPA = "IPA"
    IPL_1_0 = "IPL-1.0"
    ISC = "ISC"
    JasPer_2_0 = "JasPer-2.0"
    JPNIC = "JPNIC"
    JSON = "JSON"
    LAL_1_2 = "LAL-1.2"
    LAL_1_3 = "LAL-1.3"
    Latex2e = "Latex2e"
    Leptonica = "Leptonica"
    LGPL_2_0 = "LGPL-2.0"
    LGPL_2_0_only = "LGPL-2.0-only"
    LGPL_2_0_or_later = "LGPL-2.0-or-later"
    LGPL_2_0_ = "LGPL-2.0+"
    LGPL_2_1 = "LGPL-2.1"
    LGPL_2_1_only = "LGPL-2.1-only"
    LGPL_2_1_or_later = "LGPL-2.1-or-later"
    LGPL_2_1_ = "LGPL-2.1+"
    LGPL_3_0 = "LGPL-3.0"
    LGPL_3_0_only = "LGPL-3.0-only"
    LGPL_3_0_or_later = "LGPL-3.0-or-later"
    LGPL_3_0_ = "LGPL-3.0+"
    LGPLLR = "LGPLLR"
    Libpng = "Libpng"
    libpng_2_0 = "libpng-2.0"
    libselinux_1_0 = "libselinux-1.0"
    libtiff = "libtiff"
    LiLiQ_P_1_1 = "LiLiQ-P-1.1"
    LiLiQ_R_1_1 = "LiLiQ-R-1.1"
    LiLiQ_Rplus_1_1 = "LiLiQ-Rplus-1.1"
    Linux_OpenIB = "Linux-OpenIB"
    LPL_1_0 = "LPL-1.0"
    LPL_1_02 = "LPL-1.02"
    LPPL_1_0 = "LPPL-1.0"
    LPPL_1_1 = "LPPL-1.1"
    LPPL_1_2 = "LPPL-1.2"
    LPPL_1_3a = "LPPL-1.3a"
    LPPL_1_3c = "LPPL-1.3c"
    MakeIndex = "MakeIndex"
    MirOS = "MirOS"
    MIT = "MIT"
    MIT_0 = "MIT-0"
    MIT_advertising = "MIT-advertising"
    MIT_CMU = "MIT-CMU"
    MIT_enna = "MIT-enna"
    MIT_feh = "MIT-feh"
    MIT_Modern_Variant = "MIT-Modern-Variant"
    MIT_open_group = "MIT-open-group"
    MITNFA = "MITNFA"
    Motosoto = "Motosoto"
    mpich2 = "mpich2"
    MPL_1_0 = "MPL-1.0"
    MPL_1_1 = "MPL-1.1"
    MPL_2_0 = "MPL-2.0"
    MPL_2_0_no_copyleft_exception = "MPL-2.0-no-copyleft-exception"
    MS_PL = "MS-PL"
    MS_RL = "MS-RL"
    MTLL = "MTLL"
    MulanPSL_1_0 = "MulanPSL-1.0"
    MulanPSL_2_0 = "MulanPSL-2.0"
    Multics = "Multics"
    Mup = "Mup"
    NAIST_2003 = "NAIST-2003"
    NASA_1_3 = "NASA-1.3"
    Naumen = "Naumen"
    NBPL_1_0 = "NBPL-1.0"
    NCGL_UK_2_0 = "NCGL-UK-2.0"
    NCSA = "NCSA"
    Net_SNMP = "Net-SNMP"
    NetCDF = "NetCDF"
    Newsletr = "Newsletr"
    NGPL = "NGPL"
    NIST_PD = "NIST-PD"
    NIST_PD_fallback = "NIST-PD-fallback"
    NLOD_1_0 = "NLOD-1.0"
    NLPL = "NLPL"
    Nokia = "Nokia"
    NOSL = "NOSL"
    Noweb = "Noweb"
    NPL_1_0 = "NPL-1.0"
    NPL_1_1 = "NPL-1.1"
    NPOSL_3_0 = "NPOSL-3.0"
    NRL = "NRL"
    NTP = "NTP"
    NTP_0 = "NTP-0"
    Nunit = "Nunit"
    O_UDA_1_0 = "O-UDA-1.0"
    OCCT_PL = "OCCT-PL"
    OCLC_2_0 = "OCLC-2.0"
    ODbL_1_0 = "ODbL-1.0"
    ODC_By_1_0 = "ODC-By-1.0"
    OFL_1_0 = "OFL-1.0"
    OFL_1_0_no_RFN = "OFL-1.0-no-RFN"
    OFL_1_0_RFN = "OFL-1.0-RFN"
    OFL_1_1 = "OFL-1.1"
    OFL_1_1_no_RFN = "OFL-1.1-no-RFN"
    OFL_1_1_RFN = "OFL-1.1-RFN"
    OGC_1_0 = "OGC-1.0"
    OGDL_Taiwan_1_0 = "OGDL-Taiwan-1.0"
    OGL_Canada_2_0 = "OGL-Canada-2.0"
    OGL_UK_1_0 = "OGL-UK-1.0"
    OGL_UK_2_0 = "OGL-UK-2.0"
    OGL_UK_3_0 = "OGL-UK-3.0"
    OGTSL = "OGTSL"
    OLDAP_1_1 = "OLDAP-1.1"
    OLDAP_1_2 = "OLDAP-1.2"
    OLDAP_1_3 = "OLDAP-1.3"
    OLDAP_1_4 = "OLDAP-1.4"
    OLDAP_2_0 = "OLDAP-2.0"
    OLDAP_2_0_1 = "OLDAP-2.0.1"
    OLDAP_2_1 = "OLDAP-2.1"
    OLDAP_2_2 = "OLDAP-2.2"
    OLDAP_2_2_1 = "OLDAP-2.2.1"
    OLDAP_2_2_2 = "OLDAP-2.2.2"
    OLDAP_2_3 = "OLDAP-2.3"
    OLDAP_2_4 = "OLDAP-2.4"
    OLDAP_2_5 = "OLDAP-2.5"
    OLDAP_2_6 = "OLDAP-2.6"
    OLDAP_2_7 = "OLDAP-2.7"
    OLDAP_2_8 = "OLDAP-2.8"
    OML = "OML"
    OpenSSL = "OpenSSL"
    OPL_1_0 = "OPL-1.0"
    OSET_PL_2_1 = "OSET-PL-2.1"
    OSL_1_0 = "OSL-1.0"
    OSL_1_1 = "OSL-1.1"
    OSL_2_0 = "OSL-2.0"
    OSL_2_1 = "OSL-2.1"
    OSL_3_0 = "OSL-3.0"
    Parity_6_0_0 = "Parity-6.0.0"
    Parity_7_0_0 = "Parity-7.0.0"
    PDDL_1_0 = "PDDL-1.0"
    PHP_3_0 = "PHP-3.0"
    PHP_3_01 = "PHP-3.01"
    Plexus = "Plexus"
    PolyForm_Noncommercial_1_0_0 = "PolyForm-Noncommercial-1.0.0"
    PolyForm_Small_Business_1_0_0 = "PolyForm-Small-Business-1.0.0"
    PostgreSQL = "PostgreSQL"
    PSF_2_0 = "PSF-2.0"
    psfrag = "psfrag"
    psutils = "psutils"
    Python_2_0 = "Python-2.0"
    Qhull = "Qhull"
    QPL_1_0 = "QPL-1.0"
    Rdisc = "Rdisc"
    RHeCos_1_1 = "RHeCos-1.1"
    RPL_1_1 = "RPL-1.1"
    RPL_1_5 = "RPL-1.5"
    RPSL_1_0 = "RPSL-1.0"
    RSA_MD = "RSA-MD"
    RSCPL = "RSCPL"
    Ruby = "Ruby"
    SAX_PD = "SAX-PD"
    Saxpath = "Saxpath"
    SCEA = "SCEA"
    Sendmail = "Sendmail"
    Sendmail_8_23 = "Sendmail-8.23"
    SGI_B_1_0 = "SGI-B-1.0"
    SGI_B_1_1 = "SGI-B-1.1"
    SGI_B_2_0 = "SGI-B-2.0"
    SHL_0_5 = "SHL-0.5"
    SHL_0_51 = "SHL-0.51"
    SimPL_2_0 = "SimPL-2.0"
    SISSL = "SISSL"
    SISSL_1_2 = "SISSL-1.2"
    Sleepycat = "Sleepycat"
    SMLNJ = "SMLNJ"
    SMPPL = "SMPPL"
    SNIA = "SNIA"
    Spencer_86 = "Spencer-86"
    Spencer_94 = "Spencer-94"
    Spencer_99 = "Spencer-99"
    SPL_1_0 = "SPL-1.0"
    SSH_OpenSSH = "SSH-OpenSSH"
    SSH_short = "SSH-short"
    SSPL_1_0 = "SSPL-1.0"
    StandardML_NJ = "StandardML-NJ"
    SugarCRM_1_1_3 = "SugarCRM-1.1.3"
    SWL = "SWL"
    TAPR_OHL_1_0 = "TAPR-OHL-1.0"
    TCL = "TCL"
    TCP_wrappers = "TCP-wrappers"
    TMate = "TMate"
    TORQUE_1_1 = "TORQUE-1.1"
    TOSL = "TOSL"
    TU_Berlin_1_0 = "TU-Berlin-1.0"
    TU_Berlin_2_0 = "TU-Berlin-2.0"
    UCL_1_0 = "UCL-1.0"
    Unicode_DFS_2015 = "Unicode-DFS-2015"
    Unicode_DFS_2016 = "Unicode-DFS-2016"
    Unicode_TOU = "Unicode-TOU"
    Unlicense = "Unlicense"
    UPL_1_0 = "UPL-1.0"
    Vim = "Vim"
    VOSTROM = "VOSTROM"
    VSL_1_0 = "VSL-1.0"
    W3C = "W3C"
    W3C_19980720 = "W3C-19980720"
    W3C_20150513 = "W3C-20150513"
    Watcom_1_0 = "Watcom-1.0"
    Wsuipa = "Wsuipa"
    WTFPL = "WTFPL"
    wxWindows = "wxWindows"
    X11 = "X11"
    Xerox = "Xerox"
    XFree86_1_1 = "XFree86-1.1"
    xinetd = "xinetd"
    Xnet = "Xnet"
    xpp = "xpp"
    XSkat = "XSkat"
    YPL_1_0 = "YPL-1.0"
    YPL_1_1 = "YPL-1.1"
    Zed = "Zed"
    Zend_2_0 = "Zend-2.0"
    Zimbra_1_3 = "Zimbra-1.3"
    Zimbra_1_4 = "Zimbra-1.4"
    Zlib = "Zlib"
    zlib_acknowledgement = "zlib-acknowledgement"
    ZPL_1_1 = "ZPL-1.1"
    ZPL_2_0 = "ZPL-2.0"
    ZPL_2_1 = "ZPL-2.1"


class ContributionTypeEnum(Enum):
    """Contribution type using emojis from https://allcontributors.org/docs/en/emoji-key ."""

    audio = "audio"
    ally = "ally"
    bug = "bug"
    blog = "blog"
    business = "business"
    code = "code"
    content = "content"
    data = "data"
    doc = "doc"
    design = "design"
    example = "example"
    eventOrganizing = "eventOrganizing"
    financial = "financial"
    fundingFinding = "fundingFinding"
    ideas = "ideas"
    infra = "infra"
    maintenance = "maintenance"
    mentoring = "mentoring"
    platform = "platform"
    plugin = "plugin"
    projectManagement = "projectManagement"
    promotion = "promotion"
    question = "question"
    research = "research"
    review = "review"
    security = "security"
    tool = "tool"
    translation = "translation"
    test = "test"
    tutorial = "tutorial"
    talk = "talk"
    userTesting = "userTesting"
    video = "video"


class Country(Enum):
    """Country codes from https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2 . It is used for the country of a person in project metadata."""

    AD = "AD"
    AE = "AE"
    AF = "AF"
    AG = "AG"
    AI = "AI"
    AL = "AL"
    AM = "AM"
    AO = "AO"
    AQ = "AQ"
    AR = "AR"
    AS = "AS"
    AT = "AT"
    AU = "AU"
    AW = "AW"
    AX = "AX"
    AZ = "AZ"
    BA = "BA"
    BB = "BB"
    BD = "BD"
    BE = "BE"
    BF = "BF"
    BG = "BG"
    BH = "BH"
    BI = "BI"
    BJ = "BJ"
    BL = "BL"
    BM = "BM"
    BN = "BN"
    BO = "BO"
    BQ = "BQ"
    BR = "BR"
    BS = "BS"
    BT = "BT"
    BV = "BV"
    BW = "BW"
    BY = "BY"
    BZ = "BZ"
    CA = "CA"
    CC = "CC"
    CD = "CD"
    CF = "CF"
    CG = "CG"
    CH = "CH"
    CI = "CI"
    CK = "CK"
    CL = "CL"
    CM = "CM"
    CN = "CN"
    CO = "CO"
    CR = "CR"
    CU = "CU"
    CV = "CV"
    CW = "CW"
    CX = "CX"
    CY = "CY"
    CZ = "CZ"
    DE = "DE"
    DJ = "DJ"
    DK = "DK"
    DM = "DM"
    DO = "DO"
    DZ = "DZ"
    EC = "EC"
    EE = "EE"
    EG = "EG"
    EH = "EH"
    ER = "ER"
    ES = "ES"
    ET = "ET"
    FI = "FI"
    FJ = "FJ"
    FK = "FK"
    FM = "FM"
    FO = "FO"
    FR = "FR"
    GA = "GA"
    GB = "GB"
    GD = "GD"
    GE = "GE"
    GF = "GF"
    GG = "GG"
    GH = "GH"
    GI = "GI"
    GL = "GL"
    GM = "GM"
    GN = "GN"
    GP = "GP"
    GQ = "GQ"
    GR = "GR"
    GS = "GS"
    GT = "GT"
    GU = "GU"
    GW = "GW"
    GY = "GY"
    HK = "HK"
    HM = "HM"
    HN = "HN"
    HR = "HR"
    HT = "HT"
    HU = "HU"
    ID = "ID"
    IE = "IE"
    IL = "IL"
    IM = "IM"
    IN = "IN"
    IO = "IO"
    IQ = "IQ"
    IR = "IR"
    IS = "IS"
    IT = "IT"
    JE = "JE"
    JM = "JM"
    JO = "JO"
    JP = "JP"
    KE = "KE"
    KG = "KG"
    KH = "KH"
    KI = "KI"
    KM = "KM"
    KN = "KN"
    KP = "KP"
    KR = "KR"
    KW = "KW"
    KY = "KY"
    KZ = "KZ"
    LA = "LA"
    LB = "LB"
    LC = "LC"
    LI = "LI"
    LK = "LK"
    LR = "LR"
    LS = "LS"
    LT = "LT"
    LU = "LU"
    LV = "LV"
    LY = "LY"
    MA = "MA"
    MC = "MC"
    MD = "MD"
    ME = "ME"
    MF = "MF"
    MG = "MG"
    MH = "MH"
    MK = "MK"
    ML = "ML"
    MM = "MM"
    MN = "MN"
    MO = "MO"
    MP = "MP"
    MQ = "MQ"
    MR = "MR"
    MS = "MS"
    MT = "MT"
    MU = "MU"
    MV = "MV"
    MW = "MW"
    MX = "MX"
    MY = "MY"
    MZ = "MZ"
    NA = "NA"
    NC = "NC"
    NE = "NE"
    NF = "NF"
    NG = "NG"
    NI = "NI"
    NL = "NL"
    NO = "NO"
    NP = "NP"
    NR = "NR"
    NU = "NU"
    NZ = "NZ"
    OM = "OM"
    PA = "PA"
    PE = "PE"
    PF = "PF"
    PG = "PG"
    PH = "PH"
    PK = "PK"
    PL = "PL"
    PM = "PM"
    PN = "PN"
    PR = "PR"
    PS = "PS"
    PT = "PT"
    PW = "PW"
    PY = "PY"
    QA = "QA"
    RE = "RE"
    RO = "RO"
    RS = "RS"
    RU = "RU"
    RW = "RW"
    SA = "SA"
    SB = "SB"
    SC = "SC"
    SD = "SD"
    SE = "SE"
    SG = "SG"
    SH = "SH"
    SI = "SI"
    SJ = "SJ"
    SK = "SK"
    SL = "SL"
    SM = "SM"
    SN = "SN"
    SO = "SO"
    SR = "SR"
    SS = "SS"
    ST = "ST"
    SV = "SV"
    SX = "SX"
    SY = "SY"
    SZ = "SZ"
    TC = "TC"
    TD = "TD"
    TF = "TF"
    TG = "TG"
    TH = "TH"
    TJ = "TJ"
    TK = "TK"
    TL = "TL"
    TM = "TM"
    TN = "TN"
    TO = "TO"
    TR = "TR"
    TT = "TT"
    TV = "TV"
    TW = "TW"
    TZ = "TZ"
    UA = "UA"
    UG = "UG"
    UM = "UM"
    US = "US"
    UY = "UY"
    UZ = "UZ"
    VA = "VA"
    VC = "VC"
    VE = "VE"
    VG = "VG"
    VI = "VI"
    VN = "VN"
    VU = "VU"
    WF = "WF"
    WS = "WS"
    YE = "YE"
    YT = "YT"
    ZA = "ZA"
    ZM = "ZM"
    ZW = "ZW"



"""Integration with codemetapy (to re-generate codemeta as part of somesy sync)."""
import contextlib
import logging
from pathlib import Path

from importlib_metadata import version

from ..core.models import SomesyConfig
from .exec import gen_codemeta
from .utils import cff_codemeta_tempfile, update_codemeta_file

log = logging.getLogger("somesy")


def patch_codemetapy():
    """Monkey-patch codemetapy (2.5.0 -> 2.5.1)."""
    # TODO: remove once codemeta update is published)
    if version("codemetapy") != "2.5.0":
        return
    from codemeta.parsers import python as cmpy

    # https://github.com/proycon/codemetapy/blob/88098dc638e4cdfed9de6ad98002e16dfeede952/codemeta/parsers/python.py
    def fixed_metadata_from_pyproject(pyproject):
        """Parse metadata from pyproject.toml."""
        if pyproject.project and "name" in pyproject.project:
            return pyproject.project
        elif "poetry" in pyproject.tool:
            return pyproject.tool["poetry"]
        elif pyproject.tool and "name" in list(pyproject.tool.values())[0]:
            # fallback: no poetry but another tool that defines at least a name
            return list(pyproject.tool.values())[0]
        return None

    cmpy.metadata_from_pyproject = fixed_metadata_from_pyproject
    log.debug("monkeypatch codemetapy 2.5.0 -> 2.5.1")


def collect_cm_sources(conf: SomesyConfig):
    """Assemble list of inputs for codemetapy based on somesy config.

    Returns files that are supported by both somesy and codemetapy and are enabled for somesy.
    """
    cm_sources = []
    if (
        not conf.no_sync_pyproject
        and conf.pyproject_file is not None
        and conf.pyproject_file.is_file()
    ):
        cm_sources.append(conf.pyproject_file)
    # NOTE: we don't add CFF directly, because it must be handled separately
    # NOTE: add other suitable somesy targets / codemeta sources (except CFF and codemeta) here
    if (
        conf.no_sync_package_json
        and conf.package_json_file is not None
        and conf.package_json_file.is_file()
    ):
        cm_sources.append(conf.package_json_file)
    return cm_sources


def update_codemeta(conf: SomesyConfig) -> bool:
    """Generate or update codemeta file based on sources that somesy supports.

    Returns True if file has been written, False if it was up to date.
    """
    patch_codemetapy()
    cm_sources = collect_cm_sources(conf)

    # if cff file is given, convert it to codemeta tempfile and pass as extra input
    temp_cff_cm = contextlib.nullcontext(None)
    if not conf.no_sync_cff and conf.cff_file is not None:
        temp_cff_cm = cff_codemeta_tempfile(conf.cff_file)
        cm_sources.append(Path(temp_cff_cm.name))

    # run codemetapy
    with temp_cff_cm:
        cm_harvest = gen_codemeta(cm_sources)

    # check output and write file if needed
    return update_codemeta_file(conf.codemeta_file, cm_harvest)


__all__ = ["update_codemeta"]



from pathlib import Path

from typer.testing import CliRunner

from somesy.main import app

runner = CliRunner()


def test_cli_config_init(tmp_path, create_poetry_file, create_package_json):
    input_file = tmp_path / ".somesy.toml"
    input_file.write_text(Path("tests/core/data/.somesy.with_config.toml").read_text())
    cff_file = tmp_path / "CITATION.cff"
    pyproject_file = tmp_path / "pyproject.toml"
    codemeta_file = tmp_path / "codemeta.json"
    package_json_file = tmp_path / "package.json"

    create_poetry_file(pyproject_file)
    create_package_json(package_json_file)

    command_inputs = []

    # Input file path
    command_inputs.append(f"{input_file}\n")

    # no_sync_cff
    command_inputs.append("y\n")

    # CFF file path
    command_inputs.append(f"{cff_file}\n")

    # no_sync_pyproject
    command_inputs.append("y\n")

    # pyproject.toml file path
    command_inputs.append(f"{pyproject_file}\n")

    # sync_package_json
    command_inputs.append("y\n")

    # package.json file path
    command_inputs.append(f"{package_json_file}\n")

    # no_sync_codemeta
    command_inputs.append("y\n")

    # codemeta.json file path
    command_inputs.append(f"{codemeta_file}\n")

    # show_info
    command_inputs.append("\n")

    # verbose
    command_inputs.append("\n")

    # debug
    command_inputs.append("y\n")

    command_input = "".join(command_inputs)

    # test with correct inputs
    result = runner.invoke(
        app,
        [
            "init",
            "config",
        ],
        input=command_input,
    )
    assert result.exit_code == 0
    assert "Input file is updated/created at" in result.stdout
    assert f'cff_file = "{cff_file}"' in input_file.read_text()
    assert f'pyproject_file = "{pyproject_file}"' in input_file.read_text()
    assert "no_sync_cff = false" in input_file.read_text()

    # test with incorrect inputs
    pyproject_file.unlink()
    result = runner.invoke(
        app,
        [
            "init",
            "config",
        ],
        input=f"{pyproject_file}\n\n{cff_file}\n\n{pyproject_file}\n\ny\n",
    )
    assert result.exit_code == 1
    assert "No such file or directory" in result.stdout



from pathlib import Path

import pytest

from somesy.cff.writer import CFF
from somesy.core.models import LicenseEnum, Person, ProjectMetadata, SomesyInput


def test_load(tmp_path):
    not_exist_path = Path("reject/CITATION.cff")
    with pytest.raises(FileNotFoundError):
        CFF(not_exist_path, create_if_not_exists=False)

    file_path = tmp_path / "CITATION.cff"
    CFF(file_path, create_if_not_exists=True)
    assert file_path.is_file()

    reject_path = Path("tests/cff/data/reject.cff")
    with pytest.raises(ValueError):
        CFF(reject_path)


@pytest.fixture
def cff():
    return CFF(Path("tests/cff/data/CITATION.cff"))


@pytest.fixture
def base_person():
    people = {
        "family-names": "Soylu",
        "given-names": "Mustafa",
        "email": "m.soylu@fz-juelich.de",
        "orcid": "https://orcid.org/0000-0003-2637-0432",
    }
    return Person(**people)


@pytest.fixture
def new_person():
    people = {
        "family-names": "BB",
        "given-names": "AA",
        "email": "test@test.teset",
        "orcid": "https://orcid.org/0000-0001-2345-6789",
    }
    return Person(**people)


@pytest.fixture
def project_metadata():
    return SomesyInput.from_input_file(
        Path("tests/cff/data/pyproject.base.toml")
    ).project


def test_name(cff):
    # test existing name
    assert cff.name == "somesy"


def test_version(cff):
    # test existing version
    assert cff.version == "0.0.1"

    # empty entry
    cff.version = None
    assert cff.version == "0.0.1"

    # new version
    new_version = "0.2.0"
    cff.version = new_version
    assert cff.version == new_version


def test_description(cff: CFF):
    # test existing description
    assert (
        cff.description
        == "A cli tool for synchronizing CITATION.CFF with project files."
    )

    # empty entry
    cff.description = None
    assert (
        cff.description
        == "A cli tool for synchronizing CITATION.CFF with project files."
    )

    # new description
    new_description = "new description"
    cff.description = new_description
    assert cff.description == new_description


def test_keywords(cff):
    # test existing keywords
    assert cff.keywords == ["metadata", "rdm", "standards", "FAIR", "python3"]

    # empty entry
    cff.keywords = None
    assert cff.keywords == ["metadata", "rdm", "standards", "FAIR", "python3"]

    # new keywords
    new_keywords = ["new keyword"]
    cff.keywords = new_keywords
    assert cff.keywords == new_keywords


def test_license(cff):
    # test existing license
    assert cff.license == "MIT"

    # empty entry
    cff.license = None
    assert cff.license == "MIT"

    # new license
    new_license = "GPT-3"
    cff.license = new_license
    assert cff.license == new_license


def test_repository(cff):
    # test existing repository
    assert (
        cff.repository
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )

    # empty entry
    cff.repository = None
    assert (
        cff.repository
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )

    # new repository
    new_repository = "https://test.test"
    cff.repository = new_repository
    assert cff.repository == new_repository


def test_homepage(cff):
    # test existing homepage
    assert (
        cff.homepage
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )

    # empty entry
    cff.homepage = None
    assert (
        cff.homepage
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )

    # new homepage
    new_homepage = "https://test.test"
    cff.homepage = new_homepage
    assert cff.homepage == new_homepage


def test_sync(cff, project_metadata):
    cff.sync(project_metadata)
    assert cff.version == "0.1.0"


def test_authors(cff, base_person, new_person):
    # test existing authors
    assert cff.authors == [CFF._from_person(base_person)]

    # new authors
    cff.authors = [new_person]
    assert cff.authors == [CFF._from_person(new_person)]


def test_maintainers(cff, base_person, new_person):
    # test existing maintainers
    assert cff.maintainers == [CFF._from_person(base_person)]

    # new maintainers
    cff.maintainers = [new_person]
    assert cff.maintainers == [CFF._from_person(new_person)]


def test_save(tmp_path):
    # test save with default path
    file_path = tmp_path / "CITATION.cff"
    cff = CFF(file_path, create_if_not_exists=True)
    cff.save()
    assert file_path.is_file()

    # test save with custom path
    custom_path = tmp_path / "custom.cff"
    cff.save(custom_path)


@pytest.fixture
def person():
    p = {
        "given-names": "Jane",
        "email": "j.doe@example.com",
        "family-names": "Doe",
        "orcid": "https://orcid.org/0123-4567-8910",
    }
    ret = Person(**p)
    ret.set_key_order(list(p.keys()))  # custom order!
    return ret


def test_from_to_person(person):
    assert CFF._from_person(person) == person.dict(by_alias=True)

    p = CFF._to_person(CFF._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email
    assert p.orcid == person.orcid


def test_person_merge(tmp_path, person):
    def to_cff_keys(lst):
        return list(map(lambda s: s.replace("_", "-"), lst))

    cff_path = tmp_path / "CITATION.cff"
    cff = CFF(cff_path, create_if_not_exists=True)

    pm = ProjectMetadata(
        name="My awesome project",
        description="Project description",
        license=LicenseEnum.MIT,
        people=[person.copy(update=dict(author=True))],
    )
    cff.sync(pm)
    cff.save()

    # check that serialization preserves key order
    # (load raw dict from yaml and see order of keys)
    dct = cff._yaml.load(open(cff_path))
    assert list(dct["authors"][0].keys()) == to_cff_keys(person._key_order)

    # jane becomes john -> modified person
    person1b = person.copy(update={"given_names": "John", "author": True})

    # different Jane Doe with different orcid -> new person
    person2 = person.copy(
        update={
            "orcid": "https://orcid.org/4321-0987-3231",
            "email": "i.am.jane@doe.com",
            "author": True,
        }
    )
    # use different order, just for some difference
    person2.set_key_order(["given_names", "orcid", "family_names", "email"])

    # listed in "arbitrary" order in somesy metadata (new person comes first)
    pm.people = [person2, person1b]  # need to assign like that to keep _key_order
    cff.sync(pm)
    cff.save()

    # existing author order preserved
    assert cff.authors[0] == person1b.dict(by_alias=True, exclude={"author"})
    assert cff.authors[1] == person2.dict(by_alias=True, exclude={"author"})
    # existing author field order preserved
    dct = cff._yaml.load(open(cff_path, "r"))
    assert list(dct["authors"][0].keys()) == to_cff_keys(person1b._key_order)
    assert list(dct["authors"][1].keys()) == to_cff_keys(person2._key_order)

    # new person
    person3 = Person(
        **{
            "given_names": "Janice",
            "family_names": "Doethan",
            "email": "jane93@gmail.com",
            "author": True,
        }
    )
    # john has a new email address
    person1c = person1b.copy(update={"email": "john.of.us@qualityland.com"})
    # jane 2 is removed from authors, but added to maintainers
    person2.author = False
    person2.maintainer = True
    # reflect in project metadata
    pm.people = [person3, person2, person1c]
    # sync to CFF file
    cff.sync(pm)
    cff.save()

    assert len(cff.authors) == 2
    assert len(cff.maintainers) == 1
    assert cff.authors[0] == person1c.dict(by_alias=True, exclude={"author"})
    assert cff.authors[1] == person3.dict(by_alias=True, exclude={"author"})
    dct = cff._yaml.load(open(cff_path, "r"))
    assert list(dct["authors"][0].keys()) == to_cff_keys(person1c._key_order)



from pathlib import Path

import pytest

from somesy.core.models import Person, SomesyInput
from somesy.pyproject.writer import Poetry, Pyproject, SetupTools


@pytest.fixture
def poetry_path():
    return Path("tests/pyproject/data/pyproject.full.toml")


@pytest.fixture
def pyproject_poetry(poetry_path):
    return Pyproject(poetry_path)


@pytest.fixture
def setuptools_path():
    return Path("tests/pyproject/data/pyproject.setuptools.toml")


@pytest.fixture
def pyproject_setuptools(setuptools_path):
    return Pyproject(setuptools_path)


@pytest.fixture
def project_metadata(poetry_path):
    return SomesyInput.from_input_file(poetry_path).project


def test_pyproject_init_path(pyproject_poetry, poetry_path):
    # test if pyproject object is wrapped with Poetry object
    assert pyproject_poetry.path == poetry_path


def test_pyproject_init(tmp_path):
    path = tmp_path / "pyproject.toml"

    # test if it gives error when pyproject.toml file is not found
    with pytest.raises(FileNotFoundError):
        Pyproject(path)


def test_init_poetry(pyproject_poetry):
    # test if pyproject object is wrapped with Poetry object
    assert isinstance(pyproject_poetry.__wrapped__, Poetry)


def test_init_setuptools(pyproject_setuptools):
    # test if pyproject object is wrapped with Poetry object
    assert isinstance(pyproject_setuptools.__wrapped__, SetupTools)


def test_init_no_tool(tmp_path):
    file_path = tmp_path / "pyproject.toml"
    file_path.touch()

    # check if it raises error when no tool is found in pyproject.toml file
    with pytest.raises(ValueError):
        Pyproject(file_path)

    file_path.unlink()
    file_path.touch()


def test_sync(pyproject_poetry, project_metadata):
    pyproject_poetry.sync(project_metadata)

    assert pyproject_poetry.version == "0.1.0"


@pytest.fixture
def person():
    p = {
        "given-names": "John",
        "family-names": "Doe",
        "email": "test@test.test",
    }
    return Person(**p)


def test_person_to_poetry_string(person):
    poetry_string = Poetry._from_person(person)
    assert poetry_string == "John Doe <test@test.test>"


def test_person_to_setuptools_dict(person):
    setuptools_dict = SetupTools._from_person(person)
    assert setuptools_dict == {
        "name": "John Doe",
        "email": "test@test.test",
    }


def test_poetry_from_to_person(person):
    p = Poetry._to_person(Poetry._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email


def test_setuptools_from_to_person(person):
    p = SetupTools._to_person(SetupTools._from_person(person))
    assert p.full_name == person.full_name
    assert p.email == person.email



import json
from pathlib import Path

import pytest

from somesy.core.models import Person, ProjectMetadata, SomesyInput

p1 = {
    "given-names": "Jane",
    "family-names": "Doe",
    "email": "j.doe@example.com",
    "orcid": "https://orcid.org/0123-4567-8910",
}
p2 = {"given-names": "Foo", "family-names": "Bar", "email": "f.bar@example.com"}
p3 = {
    "given-names": p1["given-names"],
    "family-names": p1["family-names"],
    "email": p2["email"],
}
p4 = {**p2, "email": p1["email"]}
p5 = {**p2, "orcid": "https://orcid.org/1234-5678-9101"}
p6 = {**p5, "orcid": p1["orcid"]}


def test_same_person():
    # same is same (reflexivity)
    assert Person(**p1).same_person(Person(**p1))
    # missing orcid, different mail, different name -> not same
    assert not Person(**p1).same_person(Person(**p2))
    # missing orcid, different mail, same name -> same
    assert Person(**p1).same_person(Person(**p3))
    # missing orcid, same mail -> same
    assert Person(**p1).same_person(Person(**p4))
    # different orcid -> different
    assert not Person(**p1).same_person(Person(**p5))
    # same orcid -> same
    assert Person(**p1).same_person(Person(**p6))


def test_detect_duplicate_person():
    metadata = SomesyInput.from_input_file(Path("tests/core/data/.somesy.toml")).project

    meta = metadata.copy()
    meta.people.append(p1)
    ProjectMetadata(**meta.dict())

    # P1 ~= P3
    meta.people.append(p3)
    with pytest.raises(ValueError):
        ProjectMetadata(**meta.dict())

    # P1 ~= P4
    meta.people.pop()
    meta.people.append(p4)
    with pytest.raises(ValueError):
        ProjectMetadata(**meta.dict())

    # P1 ~= P6
    meta.people.pop()
    meta.people.append(p6)
    with pytest.raises(ValueError):
        ProjectMetadata(**meta.dict())

    # P1 /= P2
    meta.people.pop()
    meta.people.append(p2)
    ProjectMetadata(**meta.dict())

    # P1 /= P5
    meta.people.pop()
    meta.people.append(p5)
    ProjectMetadata(**meta.dict())

    # P2 ~= P5
    meta.people.append(p2)
    with pytest.raises(ValueError):
        ProjectMetadata(**meta.dict())


def test_custom_key_order():
    key_order = ["given-names", "orcid", "family-names", "email"]
    p = Person(
        **{
            "given-names": "Jane",
            "email": "mail@example.com",
            "family-names": "Doe",
        }
    )
    p.set_key_order(key_order)

    # correct subsequence of order
    expected_order = ["given_names", "family_names", "email"]
    assert list(p.dict(exclude_none=True).keys()) == expected_order
    assert list(json.loads(p.json(exclude_none=True)).keys()) == expected_order

    # added field appears in right spot
    p.orcid = "https://orcid.org/1234-5678-9101"
    assert list(p.dict(exclude_none=True).keys()) == p._key_order
    assert list(json.loads(p.json(exclude_none=True)).keys()) == p._key_order

    # fields not in key order come after all the listed ones
    p.affiliation = "Some institution"
    expected_order = p._key_order + ["affiliation"]
    assert list(p.dict(exclude_none=True).keys()) == expected_order
    assert list(json.loads(p.json(exclude_none=True)).keys()) == expected_order

    # key order also preserved by copy
    assert p.copy()._key_order == p._key_order



"""Main entry point for the somesy CLI."""
import logging
import sys

import typer

from somesy import __version__
from somesy.cli import init, sync
from somesy.core.log import SomesyLogLevel, init_log, set_log_level

app = typer.Typer()

logger = logging.getLogger("somesy")


@app.callback()
def version(value: bool):
    """Show somesy version and exit."""
    if value:
        typer.echo(f"somesy version: {__version__}")
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(
        None, "--version", help=version.__doc__, callback=version
    ),
    show_info: bool = typer.Option(
        None,
        "--info",
        "-v",
        help="Enable basic logging.",
    ),
    verbose: bool = typer.Option(
        None,
        "--verbose",
        "-vv",
        help="Enable verbose logging.",
    ),
    debug: bool = typer.Option(
        None,
        "--debug",
        "-vvv",
        help="Enable debug logging.",
    ),
):
    """General flags and arguments for somesy."""
    init_log()

    if sum(map(int, map(bool, [show_info, verbose, debug]))) > 1:
        typer.echo(
            "Only one of --info, --verbose or --debug may be set!", file=sys.stderr
        )
        raise typer.Exit(1)

    if show_info or verbose or debug:
        # NOTE: only explicitly setting log level if a flag is passed,
        # in order to distinguish from using the "default log level"
        # (needed to check if the user did override the log level as a CLI flag)
        set_log_level(
            SomesyLogLevel.from_flags(info=show_info, verbose=verbose, debug=debug)
        )


# add subcommands
app.add_typer(sync.app, name="sync")
app.add_typer(init.app, name="init")



"""Project metadata writer base-class."""
import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from somesy.core.models import Person, ProjectMetadata

log = logging.getLogger("somesy")


class ProjectMetadataWriter(ABC):
    """Base class for Project Metadata Output Wrapper.

    All supported output formats are implemented as subclasses.
    """

    def __init__(
        self,
        path: Path,
        *,
        create_if_not_exists: Optional[bool] = False,
        direct_mappings: Dict[str, List[str]] = None,
    ) -> None:
        """Initialize the Project Metadata Output Wrapper.

        Use the `direct_mappings` dict to define
        format-specific location for certain fields,
        if no additional processing is needed that
        requires a customized setter.

        Args:
            path: Path to target output file.
            create_if_not_exists: Create an empty CFF file if not exists. Defaults to True.
            direct_mappings: Dict with direct mappings of keys between somesy and target
        """
        self._data: Dict = {}
        self.path = path
        self.create_if_not_exists = create_if_not_exists
        self.direct_mappings = direct_mappings or {}

        if self.path.is_file():
            self._load()
            self._validate()
        else:
            if self.create_if_not_exists:
                self._init_new_file()
            else:
                raise FileNotFoundError(f"The file {self.path} does not exist.")

    def _init_new_file(self) -> None:
        """Create an new suitable target file.

        Override to initialize file with minimal contents, if needed.
        Make sure to set `self._data` to match the contents.
        """
        self.path.touch()

    @abstractmethod
    def _load(self):
        """Load the output file and validate it.

        Implement this method so that it loads the file `self.path`
        into the `self._data` dict.

        The file is guaranteed to exist.
        """

    @abstractmethod
    def _validate(self):
        """Validate the target file data.

        Implement this method so that it checks
        the validity of the metadata (relevant to somesy)
        in that file and raises exceptions on failure.
        """

    @abstractmethod
    def save(self, path: Optional[Path]) -> None:
        """Save the output file to the given path.

        Implement this in a way that will carefully
        update the target file with new metadata
        without destroying its other contents or structure.
        """

    def _get_property(self, key: Union[str, List[str]]) -> Optional[Any]:
        """Get a property from the data.

        Override this to e.g. rewrite the retrieved key
        (e.g. if everything relevant is in some subobject).
        """
        key_path = [key] if isinstance(key, str) else key

        curr = self._data
        for k in key_path:
            curr = curr.get(k)
            if curr is None:
                return None

        return curr

    def _set_property(self, key: Union[str, List[str]], value: Any) -> None:
        """Set a property in the data.

        Override this to e.g. rewrite the retrieved key
        (e.g. if everything relevant is in some subobject).
        """
        if not value:
            return
        key_path = [key] if isinstance(key, str) else key
        # create path on the fly if needed
        curr = self._data
        for key in key_path[:-1]:
            if key not in curr:
                curr[key] = {}
            curr = curr[key]
        curr[key_path[-1]] = value

    # ----
    # special handling for person metadata

    def _merge_person_metadata(
        self, old: List[Person], new: List[Person]
    ) -> List[Person]:
        """Update metadata of a list of persons.

        Will identify people based on orcid, email or full name.

        If old list has same person listed multiple times,
        the resulting list will too (we cannot correctly merge for external formats.)
        """
        new_people = []  # list for new people (e.g. added authors)
        # flag, meaning "person was not removed"
        still_exists = [False for i in range(len(old))]
        # copies of old person data, to be modified
        modified_people = [p.copy() for p in old]

        for person_meta in new:
            person_update = person_meta.dict()
            person_existed = False
            for i in range(len(modified_people)):
                person = modified_people[i]
                if not person.same_person(person_meta):
                    continue

                # not new person (-> will not append new record)
                person_existed = True
                # still exists (-> will not be removed from list)
                still_exists[i] = True

                # if there were changes -> update person
                overlapping_fields = person.dict(include=set(person_update.keys()))
                if person_update != overlapping_fields:
                    modified_people[i] = person.copy(update=person_update)

                    # show effective update in debug log
                    old_fmt = self._from_person(person)
                    new_fmt = self._from_person(modified_people[i])
                    if old_fmt != new_fmt:
                        log.debug(f"Updating person\n{old_fmt}\nto\n{new_fmt}")

            if not person_existed:
                new_people.append(person_meta)

        # show added and removed people in debug log
        removed_people = [old[i] for i in range(len(old)) if not still_exists[i]]
        for person in removed_people:
            pers_fmt = self._from_person(person)
            log.debug(f"Removing person\n{pers_fmt}")
        for person in new_people:
            pers_fmt = self._from_person(person)
            log.debug(f"Adding person\n{pers_fmt}")

        # return updated list of (still existing) people,
        # and all new people coming after them.
        existing_modified = [
            modified_people[i] for i in range(len(old)) if still_exists[i]
        ]
        return existing_modified + new_people

    def _sync_person_list(self, old: List[Any], new: List[Person]) -> List[Any]:
        old_people: List[Person] = self._parse_people(old)
        return self._merge_person_metadata(old_people, new)

    def sync(self, metadata: ProjectMetadata) -> None:
        """Sync output file with other metadata files."""
        self.name = metadata.name
        self.description = metadata.description

        if metadata.version:
            self.version = metadata.version

        if metadata.keywords:
            self.keywords = metadata.keywords

        self.authors = self._sync_person_list(self.authors, metadata.authors())
        self.maintainers = self._sync_person_list(
            self.maintainers, metadata.maintainers()
        )

        self.license = metadata.license.value
        if metadata.homepage:
            self.homepage = str(metadata.homepage)
        if metadata.repository:
            self.repository = str(metadata.repository)

    @staticmethod
    @abstractmethod
    def _from_person(person: Person) -> Any:
        """Convert a `Person` object into suitable target format."""

    @staticmethod
    @abstractmethod
    def _to_person(person_obj: Any) -> Person:
        """Convert an object representing a person into a `Person` object."""

    @classmethod
    def _parse_people(cls, people: Optional[List[Any]]) -> List[Person]:
        """Return a list of Persons parsed from list format-specific people representations."""
        return list(map(cls._to_person, people or []))

    # ----
    # individual magic getters and setters

    def _get_key(self, key):
        return self.direct_mappings.get(key) or key

    @property
    def name(self):
        """Return the name of the project."""
        return self._get_property(self._get_key("name"))

    @name.setter
    def name(self, name: str) -> None:
        """Set the name of the project."""
        self._set_property(self._get_key("name"), name)

    @property
    def version(self) -> Optional[str]:
        """Return the version of the project."""
        return self._get_property(self._get_key("version"))

    @version.setter
    def version(self, version: str) -> None:
        """Set the version of the project."""
        self._set_property(self._get_key("version"), version)

    @property
    def description(self) -> Optional[str]:
        """Return the description of the project."""
        return self._get_property(self._get_key("description"))

    @description.setter
    def description(self, description: str) -> None:
        """Set the description of the project."""
        self._set_property(self._get_key("description"), description)

    @property
    def authors(self):
        """Return the authors of the project."""
        return self._get_property(self._get_key("authors"))

    @authors.setter
    def authors(self, authors: List[Person]) -> None:
        """Set the authors of the project."""
        authors = [self._from_person(c) for c in authors]
        self._set_property(self._get_key("authors"), authors)

    @property
    def maintainers(self):
        """Return the maintainers of the project."""
        return self._get_property(self._get_key("maintainers"))

    @maintainers.setter
    def maintainers(self, maintainers: List[Person]) -> None:
        """Set the maintainers of the project."""
        maintainers = [self._from_person(c) for c in maintainers]
        self._set_property(self._get_key("maintainers"), maintainers)

    @property
    def keywords(self) -> Optional[List[str]]:
        """Return the keywords of the project."""
        return self._get_property(self._get_key("keywords"))

    @keywords.setter
    def keywords(self, keywords: List[str]) -> None:
        """Set the keywords of the project."""
        self._set_property(self._get_key("keywords"), keywords)

    @property
    def license(self) -> Optional[str]:
        """Return the license of the project."""
        return self._get_property(self._get_key("license"))

    @license.setter
    def license(self, license: Optional[str]) -> None:
        """Set the license of the project."""
        self._set_property(self._get_key("license"), license)

    @property
    def homepage(self) -> Optional[str]:
        """Return the homepage url of the project."""
        return self._get_property(self._get_key("homepage"))

    @homepage.setter
    def homepage(self, homepage: Optional[str]) -> None:
        """Set the homepage url of the project."""
        self._set_property(self._get_key("homepage"), homepage)

    @property
    def repository(self) -> Optional[Union[str, dict]]:
        """Return the repository url of the project."""
        return self._get_property(self._get_key("repository"))

    @repository.setter
    def repository(self, repository: Optional[Union[str, dict]]) -> None:
        """Set the repository url of the project."""
        self._set_property(self._get_key("repository"), repository)



"""Utility functions for CLI commands."""
import logging
import traceback

import typer
import wrapt

logger = logging.getLogger("somesy")


@wrapt.decorator
def wrap_exceptions(wrapped, instance, args, kwargs):
    """Format and log exceptions for cli commands."""
    try:
        return wrapped(*args, **kwargs)

    except Exception as e:
        logger.error(f"[bold red]Error: {e}[/bold red]")
        logger.debug(f"[red]{traceback.format_exc()}[/red]")
        raise typer.Exit(code=1) from e



import logging

import pytest
from typer.testing import CliRunner

from somesy.core.log import SomesyLogLevel, set_log_level
from somesy.main import app

runner = CliRunner()
logger = logging.getLogger("somesy")


def test_app_version():
    result = runner.invoke(app, ["--version"])
    assert result.exit_code == 0
    assert "somesy version: " in result.stdout


@pytest.mark.parametrize("log_level", [lv for lv in SomesyLogLevel])
def test_log_levels(log_level):
    set_log_level(log_level)
    assert logger.getEffectiveLevel() == SomesyLogLevel.to_logging(log_level)

    # print stuff to see that rich is always enabled
    # but configured matching the log level
    print(f"testing log level {log_level}")
    logger.warning("warning")
    logger.warning({"some": "dict"})
    logger.info("info")
    logger.verbose("verbose")  # type: ignore
    logger.debug("debug")



from pathlib import Path

import pytest
from pydantic import ValidationError

from somesy.pyproject.models import PoetryConfig

base_config = {
    "name": "somesy",
    "description": "desc",
    "version": "0.1.1",
    "authors": ["Mustafa Soylu <a@a.a>"],
}


@pytest.fixture
def cfg():
    return dict(base_config)


def test_name(cfg):
    # base config runs without an error
    PoetryConfig(**base_config)

    # package names with error
    cfg["name"] = ""
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["name"] = "asd::"
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)


def test_version_reject(cfg):
    # package version with error
    cfg["version"] = ""
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["version"] = "0.0.1..0.2"
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)


def test_license(cfg):
    # without an error
    cfg["license"] = "MIT"
    PoetryConfig(**base_config)

    # package license with error
    cfg["license"] = "FZJ"
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)


def test_authors(cfg):
    # authors with error
    cfg["authors"] = "Mustafa Soylu <a@a.a>"
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["authors"] = ["Mustafa Soylu a@a.a"]
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["authors"] = ["Mustafa Soylu <aa.a>"]
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)


def test_maintainers_accept(cfg):
    # without an error
    cfg["maintainers"] = ["Mustafa Soylu <a@a.a>"]
    PoetryConfig(**base_config)


def test_maintainers_reject(cfg):
    # maintainers with error
    cfg["maintainers"] = "Mustafa Soylu <a@a.a>"
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["maintainers"] = ["Mustafa Soylu a@a.a"]
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["maintainers"] = ["Mustafa Soylu <aa.a>"]
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)


def test_readme_accept(cfg):
    # without an error
    cfg["readme"] = Path("README.md")
    PoetryConfig(**base_config)

    cfg["readme"] = [Path("README.md")]
    PoetryConfig(**base_config)


def test_readme_reject(cfg):
    # readme with error
    cfg["readme"] = "Mustafa Soylu <a@a.a>"
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)

    cfg["readme"] = ["Readme2.md"]
    with pytest.raises(ValidationError):
        PoetryConfig(**cfg)



"""Wrappers around codemetapy and cffconvert Python APIs."""
import json
import logging
from contextlib import redirect_stderr
from io import StringIO
from pathlib import Path
from typing import Dict, List

from cffconvert.cli.create_citation import create_citation
from codemeta.codemeta import build
from codemeta.serializers.jsonld import serialize_to_jsonld

log = logging.getLogger("somesy")


def cff_to_codemeta(cff_file: Path) -> Dict:
    """Get codemeta LD dict from CITATION.cff via cffconvert."""
    return json.loads(create_citation(cff_file, None).as_codemeta())


def gen_codemeta(sources: List[str], *, with_entrypoints: bool = True) -> Dict:
    """Harvest codemeta LD dict via codemetapy."""
    log.debug(f"Running codemetapy with sources {sources}")
    with redirect_stderr(StringIO()) as cm_log:
        g, res, args, _ = build(
            inputsources=list(map(str, sources)),
            output="json",
            with_entrypoints=with_entrypoints,
        )
    # add captured codemetapy log into our log
    log.debug(f"codemetapy log:\n----\n{cm_log.getvalue()}\n----")

    return serialize_to_jsonld(g, res, args)



import pytest
from pydantic import ValidationError

from somesy.package_json.models import PackageJsonConfig

base_config = {
    "name": "somesy",
    "version": "0.0.1",
    "description": "A cli tool for synchronizing CITATION.CFF with project files.",
    "keywords": ["react", "javascript"],
    "author": "John Doe <john@doe.com> (http://johndoe.com/)",
    "license": "MIT",
}


@pytest.fixture
def cfg():
    return dict(base_config)


def test_init(cfg):
    # base config runs without an error
    PackageJsonConfig(**base_config)


def test_name(cfg):
    # package names with error
    cfg["name"] = ""
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)

    cfg["name"] = "asd::"
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)


def test_version_reject(cfg):
    # package version with error
    cfg["version"] = ""
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)

    cfg["version"] = "0.0.1..0.2"
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)


def test_license(cfg):
    # without an error
    cfg["license"] = "MIT"
    PackageJsonConfig(**cfg)


def test_author(cfg):
    # author with error
    cfg["author"] = "John Doe <a@a.a>"
    PackageJsonConfig(**cfg)

    cfg["author"] = "John Doe <aa.a>"
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)

    cfg["author"] = "John Doe <a@a.a> (http:/a.a)"
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)

    cfg["author"] = {"name": "John Doe", "email": "a@a.a", "url": "http://a.a"}
    PackageJsonConfig(**cfg)


def test_maintainers_accept(cfg):
    # without an error
    cfg["maintainers"] = ["John Doe <a@a.a>"]
    PackageJsonConfig(**cfg)


def test_maintainers_reject(cfg):
    # maintainers with error
    cfg["maintainers"] = "John Doe <a@a.a>"
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)

    cfg["maintainers"] = ["John Doe <aa.a>"]
    with pytest.raises(ValidationError):
        PackageJsonConfig(**cfg)



from pathlib import Path

import pytest
from pydantic import AnyUrl
from pydantic.tools import parse_obj_as

from somesy.core.models import Person
from somesy.pyproject.writer import Poetry


@pytest.fixture
def poetry():
    return Poetry(Path("tests/pyproject/data/pyproject.full.toml"))


def test_init(tmp_path):
    file_path = tmp_path / "pyproject.toml"

    # check for error if file doesn't exist
    with pytest.raises(FileNotFoundError):
        Poetry(file_path)


def test_key_error(poetry):
    assert poetry._get_property("not-existing") is None


def test_name(poetry):
    assert poetry.name == "somesy"
    poetry.name = "new name"
    assert poetry.name == "new name"


def test_version(poetry):
    assert poetry.version == "0.0.1"
    poetry.version = "0.0.2"
    assert poetry.version == "0.0.2"


def test_description(poetry):
    assert (
        poetry.description
        == "A cli tool for synchronizing CITATION.CFF with project files."
    )
    poetry.description = "new description"
    assert poetry.description == "new description"


def test_authors(poetry):
    # check existing authors
    authors = [
        Person(
            **{
                "email": "m.soylu@fz-juelich.de",
                "given-names": "Mustafa",
                "family-names": "Soylu",
            }
        )
    ]
    poetry.authors = authors
    assert set(poetry.authors) == set(["Mustafa Soylu <m.soylu@fz-juelich.de>"])

    # set authors
    new_authors = [
        Person(**{"email": "aaa@aaa.aaa", "given-names": "AA", "family-names": "BB"}),
    ]
    poetry.authors = new_authors
    assert set(poetry.authors) == set(["AA BB <aaa@aaa.aaa>"])


def test_maintainers(poetry):
    # check existing maintainers
    maintainers = [
        Person(
            **{
                "email": "m.soylu@fz-juelich.de",
                "given-names": "Mustafa",
                "family-names": "Soylu",
            }
        )
    ]
    poetry.maintainers = maintainers
    assert set(poetry.maintainers) == set(["Mustafa Soylu <m.soylu@fz-juelich.de>"])

    # set maintainers
    new_maintainers = [
        Person(**{"email": "aaa@aaa.aaa", "given-names": "AA", "family-names": "BB"}),
    ]
    poetry.maintainers = new_maintainers
    assert set(poetry.maintainers) == set(["AA BB <aaa@aaa.aaa>"])


def test_license(poetry):
    # check existing
    assert poetry.license == "MIT"

    # set new
    poetry.license = "GPT-3"
    assert poetry.license == "GPT-3"


def test_keywords(poetry):
    assert poetry.keywords == ["metadata", "rdm", "FAIR", "framework"]
    poetry.keywords = ["keyword1", "keyword2"]
    assert poetry.keywords == ["keyword1", "keyword2"]


def test_homepage(poetry):
    # as string
    assert (
        poetry.homepage
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )
    poetry.homepage = "https://test.test"
    assert poetry.homepage == "https://test.test"

    # as AnyUrl
    poetry.homepage = parse_obj_as(AnyUrl, "https://test.test2")
    assert poetry.homepage == "https://test.test2"


def test_repository(poetry):
    # as string
    assert (
        poetry.repository
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )

    poetry.repository = "https://test.test"
    assert poetry.repository == "https://test.test"

    poetry.repository = parse_obj_as(AnyUrl, "https://test.test2")
    assert poetry.repository == "https://test.test2"


def test_save(tmp_path, create_poetry_file):
    # test save with default path
    file_path = tmp_path / "pyproject.toml"
    create_poetry_file(file_path)
    poetry = Poetry(file_path)
    poetry.save()
    assert file_path.is_file()

    # test save with custom path
    custom_path = tmp_path / "custom.toml"
    poetry.save(custom_path)



"""somesy package."""
import importlib_metadata
from typing_extensions import Final

# Set version, it will use version from pyproject.toml if defined
__version__: Final[str] = importlib_metadata.version(__package__ or __name__)



"""Sync command for somesy."""
import logging
from pathlib import Path
from typing import Optional

import typer
from rich.pretty import pretty_repr

from somesy.commands import sync as sync_command
from somesy.core.core import discover_input
from somesy.core.log import SomesyLogLevel, get_log_level, set_log_level
from somesy.core.models import SomesyConfig, SomesyInput

from .util import wrap_exceptions

logger = logging.getLogger("somesy")

app = typer.Typer()


@app.callback(invoke_without_command=True)
@wrap_exceptions
def sync(
    input_file: Path = typer.Option(
        None,
        "--input-file",
        "-i",
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Somesy input file path (default: .somesy.toml)",
    ),
    no_sync_pyproject: bool = typer.Option(
        None,
        "--no-sync-pyproject",
        "-P",
        help="Do not sync pyproject.toml file (default: False)",
    ),
    pyproject_file: Path = typer.Option(
        None,
        "--pyproject-file",
        "-p",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Existing pyproject.toml file path (default: pyproject.toml)",
    ),
    no_sync_package_json: bool = typer.Option(
        None,
        "--sync-package-json",
        "-J",
        help="Do not sync package.json file (default: False)",
    ),
    package_json_file: Path = typer.Option(
        None,
        "--package-json-file",
        "-j",
        exists=True,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Existing package.json file path (default: package.json)",
    ),
    no_sync_cff: bool = typer.Option(
        None,
        "--no-sync-cff",
        "-C",
        help="Do not sync CITATION.cff file (default: False)",
    ),
    cff_file: Path = typer.Option(
        None,
        "--cff-file",
        "-c",
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="CITATION.cff file path (default: CITATION.cff)",
    ),
    no_sync_codemeta: bool = typer.Option(
        None,
        "--no-sync-codemeta",
        "-M",
        help="Do not sync codemeta.json file",
    ),
    codemeta_file: Path = typer.Option(
        None,
        "--codemeta-file",
        "-m",
        exists=False,
        file_okay=True,
        dir_okay=False,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Custom codemeta.json file path",
    ),
):
    """Sync project metadata input with metadata files."""
    # ---------------
    # config from CLI (merged with possibly set CLI flags for logging)
    passed_cli_args = {
        k: v
        for k, v in dict(
            input_file=discover_input(input_file),
            no_sync_cff=no_sync_cff,
            cff_file=cff_file,
            no_sync_pyproject=no_sync_pyproject,
            pyproject_file=pyproject_file,
            no_sync_package_json=no_sync_package_json,
            package_json_file=package_json_file,
            no_sync_codemeta=no_sync_codemeta,
            codemeta_file=codemeta_file,
        ).items()
        if v is not None
    }
    somesy_conf = SomesyConfig(**passed_cli_args)

    # cli_log_level is None if the user did not pass a log level (-> "default")
    cli_log_level: Optional[SomesyLogLevel] = get_log_level()

    if cli_log_level is not None:
        # update log level flags if cli log level was set
        somesy_conf.update_log_level(cli_log_level)

    somesy_input: SomesyInput = somesy_conf.get_input()

    if cli_log_level is None:
        # no cli log level -> set it according to the loaded configuration
        set_log_level(somesy_input.config.log_level())

    logger.debug(
        f"Combined config (Defaults + File + CLI):\n{pretty_repr(somesy_input.config)}"
    )
    # --------
    run_sync(somesy_input)


def run_sync(somesy_input: SomesyInput):
    """Write log messages and run synchronization based on passed config."""
    conf = somesy_input.config
    logger.info("[bold green]Synchronizing project metadata...[/bold green]")
    logger.info("Files to sync:")
    if not conf.no_sync_pyproject:
        logger.info(
            f"  - [italic]pyproject.toml[/italic]:\t[grey]{conf.pyproject_file}[/grey]"
        )
    if not conf.no_sync_package_json:
        logger.info(
            f"  - [italic]package.json[/italic]:\t[grey]{conf.package_json_file}[/grey]"
        )
    if not conf.no_sync_cff:
        logger.info(f"  - [italic]CITATION.cff[/italic]:\t[grey]{conf.cff_file}[/grey]")
    if not conf.no_sync_codemeta:
        logger.info(
            f"  - [italic]codemeta.json[/italic]:\t[grey]{conf.codemeta_file}[/grey]\n"
        )
    # ----
    sync_command(somesy_input)
    # ----
    logger.info("[bold green]Metadata synchronization completed.[/bold green]")



"""Core models for the somesy package."""
from __future__ import annotations

import functools
import json
from datetime import date
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import (
    AnyUrl,
    BaseModel,
    Extra,
    Field,
    PrivateAttr,
    root_validator,
    validator,
)
from rich.pretty import pretty_repr
from typing_extensions import Annotated

from .core import get_input_content
from .log import SomesyLogLevel
from .types import ContributionTypeEnum, Country, LicenseEnum

# --------
# Somesy configuration model


SOMESY_TARGETS = ["cff", "pyproject", "codemeta"]


class SomesyBaseModel(BaseModel):
    """Customized pydantic BaseModel for somesy.

    Apart from some general tweaks for better defaults,
    adds a private `_key_order` field, which is used to track the
    preferred order for serialization (usually coming from some existing input).

    It can be set on an instance using the set_key_order method,
    and is preserved by `copy()`.

    NOTE: The custom order is intended for leaf models (no further nested models),
    custom order will not work correctly across nesting layers.
    """

    class Config:
        """Pydantic config."""

        extra = Extra.forbid
        allow_population_by_field_name = True
        underscore_attrs_are_private = True
        anystr_strip_whitespace = True
        min_anystr_length = 1

    # ----
    # Key order magic

    _key_order: List[str] = PrivateAttr([])
    """List of field names (NOT aliases!) in the order they should be written in."""

    @classmethod
    @functools.lru_cache()  # compute once per class
    def _aliases(cls) -> Dict[str, str]:
        """Map back from alias field names to internal field names."""
        return {v.alias: k for k, v in cls.__fields__.items()}

    @classmethod
    def make_partial(cls, dct):
        """Construct unvalidated partial model from dict.

        Handles aliases correctly, unlike `construct`.
        """
        un_alias = cls._aliases()
        return cls.construct(**{un_alias.get(k) or k: v for k, v in dct.items()})

    def set_key_order(self, keys: List[str]):
        """Setter for custom key order used in serialization."""
        un_alias = self._aliases()
        # make sure we use the _actual_ field names
        self._key_order = list(map(lambda k: un_alias.get(k) or k, keys))

    def copy(self, *args, **kwargs):
        """Patched copy method (to preserve custom key order)."""
        ret = super().copy(*args, **kwargs)
        ret.set_key_order(list(self._key_order))
        return ret

    @staticmethod
    def _patch_kwargs_defaults(kwargs):
        for key in ["exclude_defaults", "exclude_none", "exclude_unset"]:
            if not kwargs.get(key):
                kwargs[key] = True

    def _reorder_dict(self, dct):
        """Return dict with patched key order (according to `self._key_order`).

        Keys in `dct` not listed in `self._key_order` come after all others.

        Used to patch up `dict()` and `json()`.
        """
        key_order = self._key_order or []
        existing = set(key_order).intersection(set(dct.keys()))
        key_order = [k for k in key_order if k in existing]
        key_order += list(set(dct.keys()) - set(key_order))
        return {k: dct[k] for k in key_order}

    def dict(self, *args, **kwargs):
        """Patched dict method (to preserve custom key order)."""
        self._patch_kwargs_defaults(kwargs)
        by_alias = kwargs.pop("by_alias", False)

        dct = super().dict(*args, **kwargs, by_alias=False)
        ret = self._reorder_dict(dct)

        if by_alias:
            ret = {self.__fields__[k].alias: v for k, v in ret.items()}
        return ret

    def json(self, *args, **kwargs):
        """Patched json method (to preserve custom key order)."""
        self._patch_kwargs_defaults(kwargs)
        by_alias = kwargs.pop("by_alias", False)

        # loop back json through dict to apply custom key order
        dct = json.loads(super().json(*args, **kwargs, by_alias=False))
        ret = self._reorder_dict(dct)

        if by_alias:
            ret = {self.__fields__[k].alias: v for k, v in ret.items()}
        return json.dumps(ret)


class SomesyConfig(SomesyBaseModel):
    """Pydantic model for somesy tool configuration.

    Note that all fields match CLI options, and CLI options will override the
    values declared in a somesy input file (such as `somesy.toml`).
    """

    @root_validator
    def at_least_one_target(cls, values):
        """Check that at least one output file is enabled."""
        if all(map(lambda x: values.get(f"no_sync_{x}"), SOMESY_TARGETS)):
            msg = "No sync target enabled, nothing to do. Probably this is a mistake?"
            raise ValueError(msg)

        return values

    # cli flags
    show_info: Annotated[
        bool,
        Field(
            description="Show basic information messages on run (-v flag).",
        ),
    ] = False
    verbose: Annotated[
        bool, Field(description="Show verbose messages on run (-vv flag).")
    ] = False
    debug: Annotated[
        bool, Field(description="Show debug messages on run (-vvv flag).")
    ] = False

    input_file: Annotated[
        Path, Field(description="Project metadata input file path.")
    ] = Path("somesy.toml")

    no_sync_pyproject: Annotated[
        bool, Field(description="Do not sync with pyproject.toml.")
    ] = False
    pyproject_file: Annotated[
        Path, Field(description="pyproject.toml file path.")
    ] = Path("pyproject.toml")

    no_sync_package_json: Annotated[
        bool, Field(description="Do not sync with package.json.")
    ] = False
    package_json_file: Annotated[
        Path, Field(description="package.json file path.")
    ] = Path("package.json")

    no_sync_cff: Annotated[bool, Field(description="Do not sync with CFF.")] = False
    cff_file: Annotated[Path, Field(description="CFF file path.")] = Path(
        "CITATION.cff"
    )

    no_sync_codemeta: Annotated[
        bool, Field(description="Do not sync with codemeta.json.")
    ] = False
    codemeta_file: Annotated[
        Path, Field(description="codemeta.json file path.")
    ] = Path("codemeta.json")

    def log_level(self) -> SomesyLogLevel:
        """Return log level derived from this configuration."""
        return SomesyLogLevel.from_flags(
            info=self.show_info, verbose=self.verbose, debug=self.debug
        )

    def update_log_level(self, log_level: SomesyLogLevel):
        """Update config flags according to passed log level."""
        self.show_info = log_level == SomesyLogLevel.INFO
        self.verbose = log_level == SomesyLogLevel.VERBOSE
        self.debug = log_level == SomesyLogLevel.DEBUG

    def get_input(self) -> SomesyInput:
        """Based on the somesy config, load the complete somesy input."""
        # get metadata+config from specified input file
        somesy_input = SomesyInput.from_input_file(self.input_file)
        # update input with merged config settings (cli overrides config file)
        dct: Dict[str, Any] = {}
        dct.update(somesy_input.config or {})
        dct.update(self.dict())
        somesy_input.config = SomesyConfig(**dct)
        return somesy_input


# --------
# Project metadata model (modified from CITATION.cff)


class Person(SomesyBaseModel):
    """Metadata abount a person in the context of a software project.

    This schema is based on CITATION.cff 1.2, modified and extended for the needs of somesy.
    """

    # NOTE: we rely on the defined aliases for direct CITATION.cff interoperability.

    orcid: Annotated[
        Optional[AnyUrl],
        Field(
            description="The person's ORCID url **(not required, but highly suggested)**."
        ),
    ]

    email: Annotated[
        str,
        Field(
            regex=r"^[\S]+@[\S]+\.[\S]{2,}$", description="The person's email address."
        ),
    ]

    family_names: Annotated[
        str, Field(alias="family-names", description="The person's family names.")
    ]
    given_names: Annotated[
        str, Field(alias="given-names", description="The person's given names.")
    ]

    name_particle: Annotated[
        Optional[str],
        Field(
            alias="name-particle",
            description="The person's name particle, e.g., a nobiliary particle or a preposition meaning 'of' or 'from' (for example 'von' in 'Alexander von Humboldt').",
            examples=["von"],
        ),
    ]
    name_suffix: Annotated[
        Optional[str],
        Field(
            alias="name-suffix",
            description="The person's name-suffix, e.g. 'Jr.' for Sammy Davis Jr. or 'III' for Frank Edwin Wright III.",
            examples=["Jr.", "III"],
        ),
    ]
    alias: Annotated[Optional[str], Field(description="The person's alias.")]

    affiliation: Annotated[
        Optional[str], Field(description="The person's affiliation.")
    ]

    address: Annotated[Optional[str], Field(description="The person's address.")]
    city: Annotated[Optional[str], Field(description="The person's city.")]
    country: Annotated[Optional[Country], Field(description="The person's country.")]
    fax: Annotated[Optional[str], Field(description="The person's fax number.")]
    post_code: Annotated[
        Optional[str], Field(alias="post-code", description="The person's post-code.")
    ]
    region: Annotated[Optional[str], Field(description="The person's region.")]
    tel: Annotated[Optional[str], Field(description="The person's phone number.")]

    # ----
    # somesy-specific extensions
    author: Annotated[
        bool,
        Field(
            description="Indicates whether the person is an author of the project (i.e. for citations)."
        ),
    ] = False
    maintainer: Annotated[
        bool,
        Field(
            description="Indicates whether the person is a maintainer of the project (i.e. for contact)."
        ),
    ] = False

    # NOTE: CFF 1.3 (once done) might provide ways for refined contributor description. That should be implemented here.
    contribution: Annotated[
        Optional[str],
        Field(description="Summary of how the person contributed to the project."),
    ]
    contribution_types: Annotated[
        Optional[List[ContributionTypeEnum]],
        Field(
            description="Relevant types of contributions (see https://allcontributors.org/docs/de/emoji-key).",
            min_items=1,
        ),
    ]
    contribution_begin: Annotated[
        Optional[date], Field(description="Beginning date of the contribution.")
    ]
    contribution_end: Annotated[
        Optional[date], Field(description="Ending date of the contribution.")
    ]

    # helper methods

    @property
    def full_name(self) -> str:
        """Return the full name of the person."""
        names = []

        if self.given_names:
            names.append(self.given_names)

        if self.name_particle:
            names.append(self.name_particle)

        if self.family_names:
            names.append(self.family_names)

        if self.name_suffix:
            names.append(self.name_suffix)

        return " ".join(names) if names else ""

    def same_person(self, other) -> bool:
        """Return whether two Person metadata records are about the same real person.

        Uses heuristic match based on orcid, email and name (whichever are provided).
        """
        if self.orcid is not None and other.orcid is not None:
            # having orcids is the best case, a real identifier
            return self.orcid == other.orcid

        # otherwise, try to match according to mail/name
        # sourcery skip: merge-nested-ifs
        if self.email is not None and other.email is not None:
            if self.email == other.email:
                # an email address belongs to exactly one person
                # => same email -> same person
                return True
            # otherwise, need to check name
            # (a person often has multiple email addresses)

        # no orcids, no/distinct email address
        # -> decide based on full_name (which is always present)
        return self.full_name == other.full_name


class ProjectMetadata(SomesyBaseModel):
    """Pydantic model for Project Metadata Input."""

    class Config:
        """Pydantic config."""

        extra = Extra.ignore

    @validator("people")
    def ensure_distinct_people(cls, people):
        """Make sure that no person is listed twice in the same person list."""
        for i in range(len(people)):
            for j in range(i + 1, len(people)):
                if people[i].same_person(people[j]):
                    p1 = pretty_repr(json.loads(people[i].json()))
                    p2 = pretty_repr(json.loads(people[j].json()))
                    msg = f"Same person is listed twice:\n{p1}\n{p2}"
                    raise ValueError(msg)
        return people

    @validator("people")
    def at_least_one_author(cls, people):
        """Make sure there is at least one author."""
        if not any(map(lambda p: p.author, people)):
            raise ValueError("At least one person must be an author of this project.")
        return people

    name: Annotated[str, Field(description="Project name.")]
    description: Annotated[str, Field(description="Project description.")]
    version: Annotated[Optional[str], Field(description="Project version.")]
    license: Annotated[LicenseEnum, Field(description="SPDX License string.")]

    repository: Annotated[
        Optional[AnyUrl],
        Field(description="URL of the project source code repository."),
    ] = None
    homepage: Annotated[
        Optional[AnyUrl], Field(description="URL of the project homepage.")
    ] = None

    keywords: Annotated[
        Optional[List[str]],
        Field(min_items=1, description="Keywords that describe the project."),
    ] = None

    people: Annotated[
        List[Person],
        Field(
            min_items=1, description="Project authors, maintainers and contributors."
        ),
    ]

    def authors(self):
        """Return people marked as authors."""
        return [p for p in self.people if p.author]

    def maintainers(self):
        """Return people marked as maintainers."""
        return [p for p in self.people if p.maintainer]


class SomesyInput(SomesyBaseModel):
    """The complete somesy input file (`somesy.toml`) or section (`pyproject.toml`)."""

    _origin: Optional[Path]

    project: Annotated[
        ProjectMetadata,
        Field(description="Project metadata to be used and synchronized."),
    ]
    config: Annotated[
        Optional[SomesyConfig],
        Field(description="somesy tool configuration (matches CLI flags)."),
    ]

    def is_somesy_file(self) -> bool:
        """Return whether this somesy input is from a somesy config file.

        That means, returns False if it is from pyproject.toml or package.json.
        """
        return self.is_somesy_file_path(self._origin or Path("."))

    @classmethod
    def is_somesy_file_path(cls, path: Path) -> bool:
        """Return whether the path looks like a somesy config file.

        That means, returns False if it is e.g. pyproject.toml or package.json.
        """
        return str(path).endswith("somesy.toml")

    @classmethod
    def from_input_file(cls, path: Path) -> SomesyInput:
        """Load somesy input from given file."""
        content = get_input_content(path)
        ret = SomesyInput(**content)
        ret._origin = path
        return ret



"""CLI command to initialize somesy configuration file."""
import logging
from pathlib import Path

import tomlkit

from somesy.core.core import get_input_content
from somesy.core.models import SomesyInput

logger = logging.getLogger("somesy")


def init_config(input_path: Path, options: dict) -> None:
    """Initialize somesy configuration file.

    Args:
        input_path (Path): Path to somesy file (will be created/overwritten).
        options (dict): CLI options.
    """
    logger.info(f"Updating input file ({input_path}) with CLI configurations...")

    content = get_input_content(input_path, no_unwrap=True)

    is_somesy = SomesyInput.is_somesy_file_path(input_path)
    input_file_type = "somesy" if is_somesy else "pyproject"
    msg = f"Found input file with {input_file_type} format."
    logger.verbose(msg)

    logger.debug(f"Input file content: {options}")

    if "input_file" in options:
        del options["input_file"]
    if is_somesy:
        content["config"] = options
    else:
        if "tool" not in content:
            content["tool"] = {}
        if "somesy" not in content["tool"]:
            content["tool"]["somesy"] = {}
        content["tool"]["somesy"]["config"] = options

    with open(input_path, "w") as f:
        tomlkit.dump(content, f)

    logger.info(f"Input file ({input_path}) updated.")
    logger.debug(f"Input file content: {content}")



from pathlib import Path

import pytest
from pydantic import AnyUrl
from pydantic.tools import parse_obj_as

from somesy.core.models import Person
from somesy.pyproject.writer import SetupTools


@pytest.fixture
def setuptools():
    return SetupTools(Path("tests/pyproject/data/pyproject.setuptools.toml"))


def test_init(tmp_path):
    file_path = tmp_path / "pyproject.toml"

    # check for error if file doesn't exist
    with pytest.raises(FileNotFoundError):
        SetupTools(file_path)


def test_key_error(setuptools):
    assert setuptools._get_property("not-existing") is None


def test_set_url(setuptools):
    # delete urls
    del setuptools._data["project"]["urls"]
    # set homepage
    setuptools.homepage = "https://test.test"
    assert setuptools.homepage == "https://test.test"


def test_name(setuptools):
    assert setuptools.name == "somesy"
    setuptools.name = "new name"
    assert setuptools.name == "new name"


def test_version(setuptools):
    assert setuptools.version == "0.0.1"
    setuptools.version = "0.0.2"
    assert setuptools.version == "0.0.2"


def test_description(setuptools):
    assert (
        setuptools.description
        == "A cli tool for synchronizing CITATION.CFF with project files."
    )
    setuptools.description = "new description"
    assert setuptools.description == "new description"


def test_authors(setuptools):
    # check existing authors
    authors = [
        Person(
            **{
                "email": "m.soylu@fz-juelich.de",
                "given-names": "Mustafa",
                "family-names": "Soylu",
            }
        )
    ]
    setuptools.authors = authors
    assert setuptools.authors == [
        {"name": "Mustafa Soylu", "email": "m.soylu@fz-juelich.de"}
    ]

    # set authors
    new_authors = [
        Person(**{"email": "aaa@aaa.aaa", "given-names": "AA", "family-names": "BB"}),
    ]
    setuptools.authors = new_authors
    assert setuptools.authors == [{"name": "AA BB", "email": "aaa@aaa.aaa"}]


def test_maintainers(setuptools):
    # check existing maintainers
    maintainers = [
        Person(
            **{
                "email": "m.soylu@fz-juelich.de",
                "given-names": "Mustafa",
                "family-names": "Soylu",
            }
        )
    ]
    setuptools.maintainers = maintainers
    assert setuptools.maintainers == [
        {"name": "Mustafa Soylu", "email": "m.soylu@fz-juelich.de"}
    ]

    # set maintainers
    new_maintainers = [
        Person(**{"email": "aaa@aaa.aaa", "given-names": "AA", "family-names": "BB"}),
    ]
    setuptools.maintainers = new_maintainers
    assert setuptools.maintainers == [{"name": "AA BB", "email": "aaa@aaa.aaa"}]


def test_license(setuptools):
    # check existing
    assert setuptools.license == "MIT"

    # set new
    setuptools.license = "GPT-3"
    assert setuptools.license == "GPT-3"


def test_keywords(setuptools):
    assert setuptools.keywords == ["metadata", "rdm", "FAIR", "framework"]
    setuptools.keywords = ["keyword1", "keyword2"]
    assert setuptools.keywords == ["keyword1", "keyword2"]


def test_homepage(setuptools):
    # as string
    assert (
        setuptools.homepage
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )
    setuptools.homepage = "https://test.test"
    assert setuptools.homepage == "https://test.test"

    # as AnyUrl
    setuptools.homepage = parse_obj_as(AnyUrl, "https://test.test2")
    assert setuptools.homepage == "https://test.test2"


def test_repository(setuptools):
    # as string
    assert (
        setuptools.repository
        == "https://github.com/Materials-Data-Science-and-Informatics/somesy"
    )

    setuptools.repository = "https://test.test"
    assert setuptools.repository == "https://test.test"

    setuptools.repository = parse_obj_as(AnyUrl, "https://test.test2")
    assert setuptools.repository == "https://test.test2"


def test_save(tmp_path, create_setuptools_file):
    # test save with default path
    file_path = tmp_path / "pyproject.toml"
    create_setuptools_file(file_path)
    setuptools = SetupTools(file_path)
    setuptools.save()
    assert file_path.is_file()

    # test save with custom path
    custom_path = tmp_path / "custom.toml"
    setuptools.save(custom_path)



"""Citation File Format (CFF) parser and saver."""
import json
from pathlib import Path
from typing import Optional

from cffconvert.cli.create_citation import create_citation
from ruamel.yaml import YAML

from somesy.core.models import Person
from somesy.core.writer import ProjectMetadataWriter


class CFF(ProjectMetadataWriter):
    """Citation File Format (CFF) parser and saver."""

    def __init__(
        self,
        path: Path,
        create_if_not_exists: bool = True,
    ):
        """Citation File Format (CFF) parser.

        See [somesy.core.writer.ProjectMetadataWriter.__init__][].
        """
        self._yaml = YAML()
        self._yaml.preserve_quotes = True

        mappings = {
            "name": ["title"],
            "description": ["abstract"],
            "homepage": ["url"],
            "repository": ["repository-code"],
            "maintainers": ["contact"],
        }
        super().__init__(
            path, create_if_not_exists=create_if_not_exists, direct_mappings=mappings
        )

    def _init_new_file(self):
        """Initialize new CFF file."""
        self._data = {
            "cff-version": "1.2.0",
            "message": "If you use this software, please cite it using these metadata.",
            "type": "software",
        }
        with open(self.path, "w") as f:
            self._yaml.dump(self._data, f)

    def _load(self):
        """Load the CFF file."""
        with open(self.path) as f:
            self._data = self._yaml.load(f)

    def _validate(self):
        """Validate the CFF file."""
        try:
            citation = create_citation(self.path, None)
            citation.validate()
        except ValueError as e:
            raise ValueError(f"CITATION.cff file is not valid!\n{e}") from e

    def save(self, path: Optional[Path] = None) -> None:
        """Save the CFF object to a file."""
        path = path or self.path
        self._yaml.dump(self._data, path)

    @staticmethod
    def _from_person(person: Person):
        """Convert project metadata person object to cff dict for person format."""
        json_str = person.json(
            exclude={
                "contribution",
                "contribution_types",
                "contribution_begin",
                "contribution_end",
                "author",
                "maintainer",
            },
            by_alias=True,  # e.g. family_names -> family-names, etc.
        )
        return json.loads(json_str)

    @staticmethod
    def _to_person(person_obj) -> Person:
        """Parse CFF Person to a somesy Person."""
        # construct (partial) Person while preserving key order from YAML
        Person._aliases()
        ret = Person.make_partial(person_obj)
        ret.set_key_order(list(person_obj.keys()))
        return ret



"""Helpers to work around issue with non-deterministic serialization."""

import importlib.resources
import json
import logging
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Dict

import rdflib
import rdflib.compare

from .exec import cff_to_codemeta

log = logging.getLogger("somesy")

# assembled context (manually downloaded and combined in a JSON array)
_CM_CONTEXT_FILE = "codemeta_context_2023-04-19.json"

# load codemeta context
with importlib.resources.open_text(__package__, _CM_CONTEXT_FILE) as c:
    _CM_CONTEXT = json.load(c)

# expected URLs
_codemeta_context = set(
    [
        "https://doi.org/10.5063/schema/codemeta-2.0",
        "https://w3id.org/software-iodata",
        "https://raw.githubusercontent.com/jantman/repostatus.org/"
        "master/badges/latest/ontology.jsonld",
        "https://schema.org",
        "https://w3id.org/software-types",
    ]
)


def _localize_codemetapy_context(json):
    """Prevent rdflib external context resolution by embedding it from a file.

    The context is required to parse the JSON-LD correctly, fields with no
    context are ignored (not considered LD).
    """
    ctx = set(json.get("@context") or [])
    if not ctx:
        # probably empty or not codemeta, nothing to do
        return json

    if ctx != _codemeta_context:
        msg = f"Unexpected codemeta context: {json['@context']}. Is this really from codemetapy?"
        raise RuntimeError(msg)

    ret = dict(json)
    ret.update({"@context": _CM_CONTEXT})

    return ret


def serialize_codemeta(cm: Dict) -> str:
    """Convert JSON Dict to str (using settings like codemetapy)."""
    # using settings like in codemetapy
    return json.dumps(cm, indent=4, ensure_ascii=False, sort_keys=True)


def _graph_from_cm_dict(graph_dict):
    """Returns codemeta with localized context from a dict produced by codemetapy."""
    g = rdflib.Graph()
    expanded = json.dumps(_localize_codemetapy_context(graph_dict))
    g.parse(data=expanded, format="json-ld")
    return g


def _graph_from_cm_file(file: Path) -> rdflib.Graph:
    """Returns loaded codemeta with localized context.

    If file does not exist, returns `None` (to distinguish from existing but empty).
    """
    if file.is_file():
        with open(file, "r") as f:
            graph_dict = json.load(f)
            return _graph_from_cm_dict(graph_dict)


# ----


def update_codemeta_file(cm_file: Path, cm_dict: Dict) -> bool:
    """Update codemeta file with graph in dict if it changed.

    Returns True if the file update happened.
    """
    old = _graph_from_cm_file(cm_file) or rdflib.Graph()
    new = _graph_from_cm_dict(cm_dict)

    if not rdflib.compare.isomorphic(old, new):
        with open(cm_file, "w") as f:
            f.write(serialize_codemeta(cm_dict))
        return True
    return False


def cff_codemeta_tempfile(cff_file: Path):
    """Returns named temporary file with codemeta export of citation file."""
    cm_cff = cff_to_codemeta(cff_file)
    temp_cff_cm = NamedTemporaryFile(prefix="cff_cm_", suffix=".json")
    temp_cff_cm.write(json.dumps(cm_cff).encode("utf-8"))
    temp_cff_cm.flush()  # needed, or it won't be readable yet
    return temp_cff_cm
